#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from gi.repository import GObject, GLib, Gtk, Peas, Pluma

from .WindowHelper import WindowHelper
from .Library import Library
from .Manager import Manager
from .comment import toggle_lines

class SnippetsPlugin(GObject.Object, Pluma.WindowActivatable):
    __gtype_name__ = "SnippetsPlugin"

    window = GObject.Property(type=Pluma.Window)

    def __init__(self):
        GObject.Object.__init__(self)

        self.dlg = None

    def system_dirs(self):
        if 'XDG_DATA_DIRS' in os.environ:
            datadirs = os.environ['XDG_DATA_DIRS']
        else:
            datadirs = '/usr/local/share' + os.pathsep + '/usr/share'

        dirs = []

        for d in datadirs.split(os.pathsep):
            d = os.path.join(d, 'pluma', 'plugins', 'snippets')

            if os.path.isdir(d):
                dirs.append(d)

        dirs.append(self.plugin_info.get_data_dir())
        return dirs

    def do_activate(self):
        library = Library()
        library.add_accelerator_callback(self.accelerator_activated)

        snippetsdir = os.path.join(GLib.get_user_config_dir(), 'pluma/snippets')
        library.set_dirs(snippetsdir, self.system_dirs())

        self._helper = WindowHelper(self)

        self._helper.run(self.window)

    def do_deactivate(self):
        library = Library()
        library.remove_accelerator_callback(self.accelerator_activated)

        self._helper.stop()
        self._helper = None

    def do_update_state(self):
        self._helper.update()

    def create_configure_dialog(self):
        window = Pluma.App.get_default().get_active_window()

        if not self.dlg:
            self.dlg = Manager(self.plugin_info.get_data_dir(), window)
        else:
            self.dlg.run(window)

        return self.dlg.dlg

    def accelerator_activated(self, group, obj, keyval, mod):
        ret = False

        if self._helper:
            ret = self._helper.accelerator_activated(keyval, mod)

        return ret

# ex:ts=4:et:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import tempfile
import shutil

from gi.repository import GObject, Gio, Gdk, Gtk, GtkSource, Pluma

from .Snippet import Snippet
from .Helper import *
from .Library import *
from .Importer import *
from .Exporter import *
from .Document import Document
from .LanguageManager import get_language_manager

class Manager:
    NAME_COLUMN = 0
    SORT_COLUMN = 1
    LANG_COLUMN = 2
    SNIPPET_COLUMN = 3
    TARGET_URI = 105

    model = None
    drag_icons = ('mate-mime-application-x-tarz', 'mate-package', 'package')
    default_export_name = _('Snippets archive') + '.tar.gz'
    dragging = False
    dnd_target_list = [Gtk.TargetEntry.new('text/uri-list', 0, TARGET_URI)]

    def __init__(self, datadir, window=None):
        self.datadir = datadir
        self.snippet = None
        self.dlg = None
        self._temp_export = None
        self.snippets_doc = None
        self.manager = None
        self.default_size = None

        self.key_press_id = 0
        self.run(window)

    def get_language_snippets(self, path, name = None):
        library = Library()

        name = self.get_language(path)
        nodes = library.get_snippets(name)

        return nodes

    def add_new_snippet_node(self, parent):
        return self.model.append(parent, ('<i>' + _('Add a new snippet...') + \
                '</i>', '', None, None))

    def fill_language(self, piter, expand=True):
        # Remove all children
        child = self.model.iter_children(piter)

        while child and self.model.remove(child):
            True

        path = self.model.get_path(piter)
        nodes = self.get_language_snippets(path)
        language = self.get_language(path)

        Library().ref(language)

        if nodes:
            for node in nodes:
                self.add_snippet(piter, node)
        else:
            # Add node that tells there are no snippets currently
            self.add_new_snippet_node(piter)

        if expand:
            self.tree_view.expand_row(path, False)

    def build_model(self, force_reload = False):
        window = Pluma.App.get_default().get_active_window()

        if window:
            view = window.get_active_view()

            if not view:
                current_lang = None
            else:
                current_lang = view.get_buffer().get_language()
                source_view = self['source_view_snippet']

        else:
            current_lang = None

        tree_view = self['tree_view_snippets']
        expand = None

        if not self.model or force_reload:
            self.model = Gtk.TreeStore(str, str, GObject.Object, object)
            self.model.set_sort_column_id(self.SORT_COLUMN, Gtk.SortType.ASCENDING)
            manager = get_language_manager()
            langs = [manager.get_language(x) for x in manager.get_language_ids()]
            langs.sort(key=lambda x: x.get_name())

            piter = self.model.append(None, (_('Global'), '', None, None))
            # Add dummy node
            self.model.append(piter, ('', '', None, None))

            nm = None

            if current_lang:
                nm = current_lang.get_name()

            for lang in langs:
                name = lang.get_name()
                parent = self.model.append(None, (name, name, lang, None))

                # Add dummy node
                self.model.append(parent, ('', '', None, None))

                if (nm == name):
                    expand = parent
        else:
            if current_lang:
                piter = self.model.get_iter_first()
                nm = current_lang.get_name()

                while piter:
                    lang = self.model.get_value(piter, \
                            self.SORT_COLUMN)

                    if lang == nm:
                        expand = piter
                        break;

                    piter = self.model.iter_next(piter)

        tree_view.set_model(self.model)

        if not expand:
            expand = self.model.get_iter_first()

        tree_view.expand_row(self.model.get_path(expand), False)
        self.select_iter(expand)

    def get_cell_data_pixbuf_cb(self, column, cell, model, iter, data):
        snippet = model.get_value(iter, self.SNIPPET_COLUMN)

        if snippet and not snippet.valid:
            cell.set_property('icon-name', 'dialog-error')
        else:
            cell.set_property('icon-name', None)

        cell.set_property('xalign', 1.0)

    def get_cell_data_cb(self, column, cell, model, iter, data):
        snippet = model.get_value(iter, self.SNIPPET_COLUMN)

        cell.set_property('editable', snippet != None)
        cell.set_property('markup', model.get_value(iter, self.NAME_COLUMN))

    def on_tree_view_drag_data_get(self, widget, context, selection_data, info, time):
        gfile = Gio.file_new_for_path(self._temp_export)
        selection_data.set_uris([gfile.get_uri()])

    def on_tree_view_drag_begin(self, widget, context):
        self.dragging = True

        if self._temp_export:
              shutil.rmtree(os.path.dirname(self._temp_export))
              self._temp_export = None

        if self.dnd_name:
            Gtk.drag_set_icon_name(context, self.dnd_name, 0, 0)

        dirname = tempfile.mkdtemp()
        filename = os.path.join(dirname, self.default_export_name)

        # Generate temporary file name
        self.export_snippets(filename, False)
        self._temp_export = filename

    def on_tree_view_drag_end(self, widget, context):
        self.dragging = False

    def on_tree_view_drag_data_received(self, widget, context, x, y, selection, info, timestamp):
        uris = selection.get_uris()

        self.import_snippets(uris)

    def on_tree_view_drag_motion(self, widget, context, x, y, timestamp):
        # Return False if we are dragging
        if self.dragging:
            return False

        # Check uri target
        if not Gtk.targets_include_uri(context.targets):
            return False

        # Check action
        action = None
        if context.suggested_action == Gdk.DragAction.COPY:
            action = Gdk.DragAction.COPY
        else:
            for act in context.actions:
                if act == Gdk.DragAction.COPY:
                      action = Gdk.DragAction.COPY
                      break

        if action == Gdk.DragAction.COPY:
            context.drag_status(Gdk.DragAction.COPY, timestamp)
            return True
        else:
            return False

    def build_dnd(self):
        tv = self.tree_view

        # Set it as a drag source for exporting snippets
        tv.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, self.dnd_target_list, Gdk.DragAction.DEFAULT | Gdk.DragAction.COPY)

        # Set it as a drag destination for importing snippets
        tv.drag_dest_set(Gtk.DestDefaults.HIGHLIGHT | Gtk.DestDefaults.DROP,
                 self.dnd_target_list, Gdk.DragAction.DEFAULT | Gdk.DragAction.COPY)

        tv.connect('drag_data_get', self.on_tree_view_drag_data_get)
        tv.connect('drag_begin', self.on_tree_view_drag_begin)
        tv.connect('drag_end', self.on_tree_view_drag_end)
        tv.connect('drag_data_received', self.on_tree_view_drag_data_received)
        tv.connect('drag_motion', self.on_tree_view_drag_motion)

        theme = Gtk.IconTheme.get_for_screen(tv.get_screen())

        self.dnd_name = None
        for name in self.drag_icons:
            icon = theme.lookup_icon(name, Gtk.IconSize.DND, 0)

            if icon:
                self.dnd_name = name
                break

    def build_tree_view(self):
        self.tree_view = self['tree_view_snippets']

        self.column = Gtk.TreeViewColumn(None)

        self.renderer = Gtk.CellRendererText()
        self.column.pack_start(self.renderer, False)
        self.column.set_cell_data_func(self.renderer, self.get_cell_data_cb, None)

        renderer = Gtk.CellRendererPixbuf()
        self.column.pack_start(renderer, True)
        self.column.set_cell_data_func(renderer, self.get_cell_data_pixbuf_cb, None)

        self.tree_view.append_column(self.column)

        self.renderer.connect('edited', self.on_cell_edited)
        self.renderer.connect('editing-started', self.on_cell_editing_started)

        selection = self.tree_view.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        selection.connect('changed', self.on_tree_view_selection_changed)

        self.build_dnd()

    def build(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(os.path.join(self.datadir, 'ui', 'snippets.ui'))

        handlers_dic = {
            'on_dialog_snippets_response': self.on_dialog_snippets_response,
            'on_dialog_snippets_destroy': self.on_dialog_snippets_destroy,
            'on_button_new_snippet_clicked': self.on_button_new_snippet_clicked,
            'on_button_import_snippets_clicked': self.on_button_import_snippets_clicked,
            'on_button_export_snippets_clicked': self.on_button_export_snippets_clicked,
            'on_button_remove_snippet_clicked': self.on_button_remove_snippet_clicked,
            'on_entry_tab_trigger_focus_out': self.on_entry_tab_trigger_focus_out,
            'on_entry_tab_trigger_changed': self.on_entry_tab_trigger_changed,
            'on_entry_accelerator_focus_out': self.on_entry_accelerator_focus_out,
            'on_entry_accelerator_focus_in': self.on_entry_accelerator_focus_in,
            'on_entry_accelerator_key_press': self.on_entry_accelerator_key_press,
            'on_source_view_snippet_focus_out': self.on_source_view_snippet_focus_out,
            'on_tree_view_snippets_row_expanded': self.on_tree_view_snippets_row_expanded,
            'on_tree_view_snippets_key_press': self.on_tree_view_snippets_key_press}

        self.builder.connect_signals(handlers_dic)

        self.build_tree_view()
        self.build_model()

        source_view = self['source_view_snippet']
        manager = get_language_manager()
        lang = manager.get_language('snippets')

        if lang:
            source_view.get_buffer().set_highlight_syntax(True)
            source_view.get_buffer().set_language(lang)
            self.snippets_doc = Document(None, source_view)

        combo = self['combo_drop_targets']

        entry = combo.get_child()
        entry.connect('focus-out-event', self.on_entry_drop_targets_focus_out)
        entry.connect('drag-data-received', self.on_entry_drop_targets_drag_data_received)

        lst = entry.drag_dest_get_target_list()
        lst.add_uri_targets(self.TARGET_URI)

        self.dlg = self['dialog_snippets']

        if self.default_size:
            self.dlg.set_default_size(*self.default_size)

    def __getitem__(self, key):
        return self.builder.get_object(key)

    def is_filled(self, piter):
        if not self.model.iter_has_child(piter):
            return True

        child = self.model.iter_children(piter)
        nm = self.model.get_value(child, self.NAME_COLUMN)
        lang = self.model.get_value(child, self.LANG_COLUMN)
        snippet = self.model.get_value(child, self.SNIPPET_COLUMN)

        return (lang or snippet or nm)

    def fill_if_needed(self, piter, expand=True):
        if not self.is_filled(piter):
            self.fill_language(piter, expand)

    def find_iter(self, parent, snippet):
        self.fill_if_needed(parent)
        piter = self.model.iter_children(parent)

        while (piter):
            node = self.model.get_value(piter, self.SNIPPET_COLUMN)

            if node == snippet.data:
                return piter

            piter = self.model.iter_next(piter)

        return None

    def selected_snippets_state(self):
        snippets = self.selected_snippets(False)
        override = False
        remove = False
        system = False

        for snippet in snippets:
            if not snippet:
                continue

            if snippet.is_override():
                override = True
            elif snippet.can_modify():
                remove = True
            else:
                system = True

            # No need to continue if both are found
            if override and remove:
                break

        return (override, remove, system)

    def update_buttons(self):
        button_remove = self['button_remove_snippet']
        button_new = self['button_new_snippet']
        image_remove = self['image_remove']

        button_new.set_sensitive(self.language_path != None)
        override, remove, system = self.selected_snippets_state()

        if not (override ^ remove) or system:
            button_remove.set_sensitive(False)
            image_remove.set_from_icon_name("edit-delete", Gtk.IconSize.BUTTON)
        else:
            button_remove.set_sensitive(True)

            if override:
                image_remove.set_from_icon_name("edit-undo", Gtk.IconSize.BUTTON)
                tooltip = _('Revert selected snippet')
            else:
                image_remove.set_from_icon_name("edit-delete", Gtk.IconSize.BUTTON)
                tooltip = _('Delete selected snippet')

            button_remove.set_tooltip_text(tooltip)

    def snippet_changed(self, piter = None):
        if piter:
            node = self.model.get_value(piter, self.SNIPPET_COLUMN)
            s = Snippet(node)
        else:
            s = self.snippet
            piter = self.find_iter(self.model.get_iter(self.language_path), s)

        if piter:
            nm = s.display()

            self.model.set_value(piter, self.NAME_COLUMN, nm)
            self.model.set_value(piter, self.SORT_COLUMN, nm)
            self.update_buttons()
            self.entry_tab_trigger_update_valid()

        return piter

    def add_snippet(self, parent, snippet):
        piter = self.model.append(parent, ('', '', None, snippet))

        return self.snippet_changed(piter)

    def run(self, window=None):
        if not self.dlg:
            self.build()
            self.dlg.set_transient_for(window)
            self.dlg.show()
        else:
            self.build_model()
            self.dlg.set_transient_for(window)
            self.dlg.present()

    def snippet_from_iter(self, model, piter):
        parent = model.iter_parent(piter)

        if parent:
            return model.get_value(piter, self.SNIPPET_COLUMN)
        else:
            return None

    def language_snippets(self, model, parent, as_path=False):
        self.fill_if_needed(parent, False)
        piter = model.iter_children(parent)
        snippets = []

        if not piter:
            return snippets

        while piter:
            snippet = self.snippet_from_iter(model, piter)

            if snippet:
                if as_path:
                    snippets.append(model.get_path(piter))
                else:
                    snippets.append(snippet)

            piter = model.iter_next(piter)

        return snippets

    def selected_snippets(self, include_languages=True, as_path=False):
        selection = self.tree_view.get_selection()
        (model, paths) = selection.get_selected_rows()
        snippets = []

        if paths and len(paths) != 0:
            for p in paths:
                piter = model.get_iter(p)
                parent = model.iter_parent(piter)

                if not piter:
                    continue

                if parent:
                    snippet = self.snippet_from_iter(model, piter)

                    if not snippet:
                        continue

                    if as_path:
                        snippets.append(p)
                    else:
                        snippets.append(snippet)
                elif include_languages:
                    snippets += self.language_snippets(model, piter, as_path)

        return snippets

    def selected_snippet(self):
        selection = self.tree_view.get_selection()
        (model, paths) = selection.get_selected_rows()

        if len(paths) == 1:
            piter = model.get_iter(paths[0])
            parent = model.iter_parent(piter)
            snippet = self.snippet_from_iter(model, piter)

            return parent, piter, snippet
        else:
            return None, None, None

    def selection_changed(self):
        if not self.snippet:
            sens = False

            self['entry_tab_trigger'].set_text('')
            self['entry_accelerator'].set_text('')
            buf = self['source_view_snippet'].get_buffer()
            buf.begin_not_undoable_action()
            buf.set_text('')
            buf.end_not_undoable_action()
            self['combo_drop_targets'].get_child().set_text('')

        else:
            sens = True

            self['entry_tab_trigger'].set_text(self.snippet['tag'])
            self['entry_accelerator'].set_text( \
                    self.snippet.accelerator_display())
            self['combo_drop_targets'].get_child().set_text(', '.join(self.snippet['drop-targets']))

            buf = self['source_view_snippet'].get_buffer()
            buf.begin_not_undoable_action()
            buf.set_text(self.snippet['text'])
            buf.end_not_undoable_action()


        for name in ['source_view_snippet', 'label_tab_trigger',
                'entry_tab_trigger', 'label_accelerator',
                'entry_accelerator', 'label_drop_targets',
                'combo_drop_targets']:
            self[name].set_sensitive(sens)

        self.update_buttons()

    def select_iter(self, piter, unselect=True):
        selection = self.tree_view.get_selection()

        if unselect:
            selection.unselect_all()

        selection.select_iter(piter)

        self.tree_view.scroll_to_cell(self.model.get_path(piter), None, \
            True, 0.5, 0.5)

    def get_language(self, path):
        if path.get_indices()[0] == 0:
            return None
        else:
            return self.model.get_value(self.model.get_iter(path), \
                            self.LANG_COLUMN).get_id()

    def new_snippet(self, properties=None):
        if not self.language_path:
            return None

        snippet = Library().new_snippet(self.get_language(self.language_path), properties)

        return Snippet(snippet)

    def get_dummy(self, parent):
        if not self.model.iter_n_children(parent) == 1:
            return None

        dummy = self.model.iter_children(parent)

        if not self.model.get_value(dummy, self.SNIPPET_COLUMN):
            return dummy

        return None

    def unref_languages(self):
        piter = self.model.get_iter_first()
        library = Library()

        while piter:
            if self.is_filled(piter):
                language = self.get_language(self.model.get_path(piter))
                library.save(language)

                library.unref(language)

            piter = self.model.iter_next(piter)

    # Callbacks
    def on_dialog_snippets_destroy(self, dlg):
        # Remove temporary drag export
        if self._temp_export:
              shutil.rmtree(os.path.dirname(self._temp_export))
              self._temp_export = None

        if self.snippets_doc:
            self.snippets_doc.stop()

        self.manager = None
        self.unref_languages()
        self.snippet = None
        self.model = None
        self.dlg = None

    def on_dialog_snippets_response(self, dlg, resp):

        alloc = dlg.get_allocation()
        self.default_size = [alloc.width, alloc.height]

        if resp == Gtk.ResponseType.HELP:
            Pluma.help_display(self.dlg, 'pluma', 'pluma-snippets-plugin')
            return

        self.dlg.destroy()

    def on_cell_editing_started(self, renderer, editable, path):
        piter = self.model.get_iter(path)

        if not self.model.iter_parent(piter):
            renderer.stop_editing(True)
            editable.remove_widget()
        elif isinstance(editable, Gtk.Entry):
            if self.snippet:
                editable.set_text(self.snippet['description'])
            else:
                # This is the `Add a new snippet...` item
                editable.set_text('')

            editable.grab_focus()

    def on_cell_edited(self, cell, path, new_text):
        if new_text != '':
            piter = self.model.get_iter(path)
            node = self.model.get_value(piter, self.SNIPPET_COLUMN)

            if node:
                if node == self.snippet.data:
                    s = self.snippet
                else:
                    s = Snippet(node)

                s['description'] = new_text
                self.snippet_changed(piter)
                self.select_iter(piter)
            else:
                # This is the `Add a new snippet...` item
                # We create a new snippet
                snippet = self.new_snippet({'description': new_text})

                if snippet:
                    self.model.set_value(piter, self.SNIPPET_COLUMN, snippet.data)
                    self.snippet_changed(piter)
                    self.snippet = snippet
                    self.selection_changed()

    def on_entry_accelerator_focus_out(self, entry, event):
        if not self.snippet:
            return

        entry.set_text(self.snippet.accelerator_display())

    def entry_tab_trigger_update_valid(self):
        entry = self['entry_tab_trigger']
        text = entry.get_text()

        if text and not Library().valid_tab_trigger(text):
            img = self['image_tab_trigger']
            img.set_from_icon_name("dialog-error", Gtk.IconSize.BUTTON)
            img.show()

            #self['hbox_tab_trigger'].set_spacing(3)
            tip = _('This is not a valid Tab trigger. Triggers can either contain letters or a single (non-alphanumeric) character like: {, [, etc.')

            entry.set_tooltip_text(tip)
            img.set_tooltip_text(tip)
        else:
            self['image_tab_trigger'].hide()
            #self['hbox_tab_trigger'].set_spacing(0)
            entry.set_tooltip_text(_('Single word the snippet is activated with after pressing Tab'))

        return False

    def on_entry_tab_trigger_focus_out(self, entry, event):
        if not self.snippet:
            return

        text = entry.get_text()

        # save tag
        self.snippet['tag'] = text
        self.snippet_changed()

    def on_entry_drop_targets_focus_out(self, entry, event):
        if not self.snippet:
            return

        text = entry.get_text()

        # save drop targets
        self.snippet['drop-targets'] = text
        self.snippet_changed()

    def on_entry_tab_trigger_changed(self, entry):
        self.entry_tab_trigger_update_valid()

    def on_source_view_snippet_focus_out(self, source_view, event):
        if not self.snippet:
            return

        buf = source_view.get_buffer()
        text = buf.get_text(buf.get_start_iter(), \
                buf.get_end_iter(), False)

        self.snippet['text'] = text
        self.snippet_changed()

    def on_button_new_snippet_clicked(self, button):
        snippet = self.new_snippet()

        if not snippet:
            return

        parent = self.model.get_iter(self.language_path)
        path = self.model.get_path(parent)

        dummy = self.get_dummy(parent)

        if dummy:
            # Remove the dummy
            self.model.remove(dummy)

        # Add the snippet
        piter = self.add_snippet(parent, snippet.data)
        self.select_iter(piter)

        if not self.tree_view.row_expanded(path):
            self.tree_view.expand_row(path, False)
            self.select_iter(piter)

        self.tree_view.grab_focus()

        path = self.model.get_path(piter)
        self.tree_view.set_cursor(path, self.column, True)

    def file_filter(self, name, pattern):
        fil = Gtk.FileFilter()
        fil.set_name(name)

        for p in pattern:
            fil.add_pattern(p)

        return fil

    def import_snippets(self, filenames):
        success = True

        for filename in filenames:
            if not Pluma.utils_uri_has_file_scheme(filename):
                continue

            # Remove file://
            gfile = Gio.file_new_for_uri(filename)
            filename = gfile.get_path()

            importer = Importer(filename)
            error = importer.run()

            if error:
                message = _('The following error occurred while importing: %s') % error
                success = False
                message_dialog(self.dlg, Gtk.MessageType.ERROR, message)

        self.build_model(True)

        if success:
            message = _('Import successfully completed')
            message_dialog(self.dlg, Gtk.MessageType.INFO, message)

    def on_import_response(self, dialog, response):
        if response == Gtk.ResponseType.CANCEL or response == Gtk.ResponseType.CLOSE:
            dialog.destroy()
            return

        f = dialog.get_uris()
        dialog.destroy()

        self.import_snippets(f)

    def on_button_import_snippets_clicked(self, button):
        dlg = Gtk.FileChooserDialog(title=_("Import snippets"),
                                    parent=self.dlg,
                                    action=Gtk.FileChooserAction.OPEN)
        self._add_button(dlg, _('_Cancel'), Gtk.ResponseType.CANCEL, "process-stop")
        self._add_button(dlg, _("_Open"), Gtk.ResponseType.OK, "document-open")

        dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar', '*.xml')))
        dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
        dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))
        dlg.add_filter(self.file_filter(_('Single snippets file'), ('*.xml',)))
        dlg.add_filter(self.file_filter(_('All files'), '*'))

        dlg.connect('response', self.on_import_response)
        dlg.set_local_only(True)

        dlg.show()

    def export_snippets_real(self, filename, snippets, show_dialogs=True):
        export = Exporter(filename, snippets)
        error = export.run()

        if error:
            message = _('The following error occurred while exporting: %s') % error
            msgtype = Gtk.MessageType.ERROR
            retval = False
        else:
            message = _('Export successfully completed')
            msgtype = Gtk.MessageType.INFO
            retval = True

        if show_dialogs:
            message_dialog(self.dlg, msgtype, message)

        return retval

    def on_export_response(self, dialog, response):
        filename = dialog.get_filename()
        snippets = dialog._export_snippets

        dialog.destroy()

        if response != Gtk.ResponseType.OK:
            return

        self.export_snippets_real(filename, snippets);

    def export_snippets(self, filename=None, show_dialogs=True):
        snippets = self.selected_snippets()

        if not snippets or len(snippets) == 0:
            return False

        usersnippets = []
        systemsnippets = []

        # Iterate through snippets and look for system snippets
        for snippet in snippets:
            if snippet.can_modify():
                usersnippets.append(snippet)
            else:
                systemsnippets.append(snippet)

        export_snippets = snippets

        if len(systemsnippets) != 0 and show_dialogs:
            # Ask if system snippets should also be exported
            message = _('Do you want to include selected <b>system</b> snippets in your export?')
            mes = Gtk.MessageDialog(flags=Gtk.DialogFlags.MODAL,
                    type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                    message_format=message)
            mes.set_property('use-markup', True)
            resp = mes.run()
            mes.destroy()

            if resp == Gtk.ResponseType.NO:
                export_snippets = usersnippets
            elif resp != Gtk.ResponseType.YES:
                return False

        if len(export_snippets) == 0 and show_dialogs:
            message = _('There are no snippets selected to be exported')
            message_dialog(self.dlg, Gtk.MessageType.INFO, message)
            return False

        if not filename:
            dlg = Gtk.FileChooserDialog(title=_('Export snippets'),
                                        parent=self.dlg,
                                        action=Gtk.FileChooserAction.SAVE)
            self._add_button(dlg, _('_Cancel'), Gtk.ResponseType.CANCEL, "process-stop")
            self._add_button(dlg, _("_Save"), Gtk.ResponseType.OK, "document-save")

            dlg._export_snippets = export_snippets
            dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar')))
            dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
            dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))

            dlg.add_filter(self.file_filter(_('All files'), '*'))
            dlg.set_do_overwrite_confirmation(True)
            dlg.set_current_name(self.default_export_name)

            dlg.connect('response', self.on_export_response)
            dlg.set_local_only(True)

            dlg.show()
            return True
        else:
            return self.export_snippets_real(filename, export_snippets, show_dialogs)

    def on_button_export_snippets_clicked(self, button):
        snippets = self.selected_snippets()

        if not snippets or len(snippets) == 0:
            return

        usersnippets = []
        systemsnippets = []

        # Iterate through snippets and look for system snippets
        for snippet in snippets:
            if snippet.can_modify():
                usersnippets.append(snippet)
            else:
                systemsnippets.append(snippet)

        dlg = Gtk.FileChooserDialog(title=_('Export snippets'),
                                    parent=self.dlg,
                                    action=Gtk.FileChooserAction.SAVE)
        self._add_button(dlg, _('_Cancel'), Gtk.ResponseType.CANCEL, "process-stop")
        self._add_button(dlg, _("_Save"), Gtk.ResponseType.OK, "document-save")

        dlg._export_snippets = snippets

        if len(systemsnippets) != 0:
            # Ask if system snippets should also be exported
            message = _('Do you want to include selected <b>system</b> snippets in your export?')
            mes = Gtk.MessageDialog(flags=Gtk.DialogFlags.MODAL,
                    type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                    message_format=message)
            mes.set_property('use-markup', True)
            resp = mes.run()
            mes.destroy()

            if resp == Gtk.ResponseType.NO:
                dlg._export_snippets = usersnippets
            elif resp != Gtk.ResponseType.YES:
                dlg.destroy()
                return

        if len(dlg._export_snippets) == 0:
            dlg.destroy()

            message = _('There are no snippets selected to be exported')
            message_dialog(self.dlg, Gtk.MessageType.INFO, message)
            return

        dlg.add_filter(self.file_filter(_('All supported archives'), ('*.gz','*.bz2','*.tar')))
        dlg.add_filter(self.file_filter(_('Gzip compressed archive'), ('*.tar.gz',)))
        dlg.add_filter(self.file_filter(_('Bzip2 compressed archive'), ('*.tar.bz2',)))

        dlg.add_filter(self.file_filter(_('All files'), '*'))
        dlg.set_do_overwrite_confirmation(True)
        dlg.set_current_name(self.default_export_name)

        dlg.connect('response', self.on_export_response)
        dlg.set_local_only(True)

        dlg.show()

    def remove_snippet_revert(self, path, piter):
        node = self.snippet_from_iter(self.model, piter)
        Library().revert_snippet(node)

        return piter

    def remove_snippet_delete(self, path, piter):
        node = self.snippet_from_iter(self.model, piter)
        parent = self.model.iter_parent(piter)

        Library().remove_snippet(node)
        idx = path.get_indices()

        if self.model.remove(piter):
            return piter
        elif idx[-1] != 0:
            self.select_iter(self.model.get_iter((idx[0], idx[1] - 1)))
        else:
            dummy = self.add_new_snippet_node(parent)
            self.tree_view.expand_row(self.model.get_path(parent), False)
            return dummy

    def on_button_remove_snippet_clicked(self, button):
        override, remove, system = self.selected_snippets_state()

        if not (override ^ remove) or system:
            return

        paths = self.selected_snippets(include_languages=False, as_path=True)

        if override:
            action = self.remove_snippet_revert
        else:
            action = self.remove_snippet_delete

        # Remove selection
        self.tree_view.get_selection().unselect_all()

        # Create tree row references
        references = []
        for path in paths:
            references.append(Gtk.TreeRowReference(self.model, path))

        # Remove/revert snippets
        select = None
        for reference in references:
            path = reference.get_path()
            piter = self.model.get_iter(path)

            res = action(path, piter)

            if res:
                select = res

        if select:
            self.select_iter(select)

        self.selection_changed()

    def set_accelerator(self, keyval, mod):
        accelerator = Gtk.accelerator_name(keyval, mod)
        self.snippet['accelerator'] = accelerator

        return True

    def on_entry_accelerator_key_press(self, entry, event):
        source_view = self['source_view_snippet']

        if event.keyval == Gdk.keyval_from_name('Escape'):
            # Reset
            entry.set_text(self.snippet.accelerator_display())
            self.tree_view.grab_focus()

            return True
        elif event.keyval == Gdk.keyval_from_name('Delete') or \
                event.keyval == Gdk.keyval_from_name('BackSpace'):
            # Remove the accelerator
            entry.set_text('')
            self.snippet['accelerator'] = ''
            self.tree_view.grab_focus()

            self.snippet_changed()
            return True
        elif Library().valid_accelerator(event.keyval, event.state):
            # New accelerator
            self.set_accelerator(event.keyval, \
                    event.state & Gtk.accelerator_get_default_mod_mask())
            entry.set_text(self.snippet.accelerator_display())
            self.snippet_changed()
            self.tree_view.grab_focus()

        else:
            return True

    def on_entry_accelerator_focus_in(self, entry, event):
        if self.snippet['accelerator']:
            entry.set_text(_('Type a new shortcut, or press Backspace to clear'))
        else:
            entry.set_text(_('Type a new shortcut'))

    def update_language_path(self):
        model, paths = self.tree_view.get_selection().get_selected_rows()

        # Check if all have the same language parent
        current_parent = None

        for path in paths:
            piter = model.get_iter(path)
            parent = model.iter_parent(piter)

            if parent:
                path = model.get_path(parent)

            if current_parent != None and current_parent != path:
                current_parent = None
                break
            else:
                current_parent = path

        self.language_path = current_parent

    def on_tree_view_selection_changed(self, selection):
        parent, piter, node = self.selected_snippet()

        if self.snippet:
            self.on_entry_tab_trigger_focus_out(self['entry_tab_trigger'],
                    None)
            self.on_source_view_snippet_focus_out(self['source_view_snippet'],
                    None)
            self.on_entry_drop_targets_focus_out(self['combo_drop_targets'].get_child(),
                    None)

        self.update_language_path()

        if node:
            self.snippet = Snippet(node)
        else:
            self.snippet = None

        self.selection_changed()

    def iter_after(self, target, after):
        if not after:
            return True

        tp = self.model.get_path(target)
        ap = self.model.get_path(after)

        if tp[0] > ap[0] or (tp[0] == ap[0] and (len(ap) == 1 or tp[1] > ap[1])):
            return True

        return False

    def on_tree_view_snippets_key_press(self, treeview, event):
        if event.keyval == Gdk.keyval_from_name('Delete'):
            self.on_button_remove_snippet_clicked(None)
            return True

    def on_tree_view_snippets_row_expanded(self, treeview, piter, path):
        # Check if it is already filled
        self.fill_if_needed(piter)
        self.select_iter(piter)

    def on_entry_drop_targets_drag_data_received(self, entry, context, x, y, selection_data, info, timestamp):
        uris = drop_get_uris(selection_data)

        if not uris:
            return

        if entry.get_text():
            mimes = [entry.get_text()]
        else:
            mimes = []

        for uri in uris:
            try:
                mime = Gio.content_type_guess(uri)
            except:
                mime = None

            if mime:
                mimes.append(mime)

        entry.set_text(', '.join(mimes))
        self.on_entry_drop_targets_focus_out(entry, None)
        context.finish(True, False, timestamp)

        entry.stop_emission('drag_data_received')

    @staticmethod
    def _add_button(dialog, label, response, icon=None):
        button = dialog.add_button(label, response)
        if icon:
            image = Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON)
            button.set_image(image)
            button.set_property("always-show-image", True)
        return button

# ex:ts=4:et:



#    Pluma snippets plugin
#    Copyright (C) 2006-2007  Jesse van den Kieboom <jesse@icecrew.nl>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re

class ParseError(Exception):
    def __str__(self):
        return 'Parse error, resume next'

class Modifiers:
    def _first_char(s):
        first = (s != '' and s[0]) or ''
        rest = (len(s) > 1 and s[1:]) or ''

        return first, rest

    def upper_first(s):
        first, rest = Modifiers._first_char(s)

        return '%s%s' % (first.upper(), rest)

    def upper(s):
        return s.upper()

    def lower_first(s):
        first, rest = Modifiers._first_char(s)

        return '%s%s' % (first.lower(), rest)

    def lower(s):
        return s.lower()

    def title(s):
        return s.title()

    upper_first = staticmethod(upper_first)
    upper = staticmethod(upper)
    lower_first = staticmethod(lower_first)
    lower = staticmethod(lower)
    title = staticmethod(title)
    _first_char = staticmethod(_first_char)

class SubstitutionParser:
    REG_ID = '[0-9]+'
    REG_NAME = '[a-zA-Z_]+'
    REG_MOD = '[a-zA-Z]+'
    REG_ESCAPE = '\\\\|\\(\\?|,|\\)'

    def __init__(self, pattern, groups = {}, modifiers = {}):
        self.pattern = pattern
        self.groups = groups

        self.REG_GROUP = '(?:(%s)|<(%s|%s)(?:,(%s))?>)' % (self.REG_ID, self.REG_ID, self.REG_NAME, self.REG_MOD)
        self.modifiers = {'u': Modifiers.upper_first,
                  'U': Modifiers.upper,
                  'l': Modifiers.lower_first,
                  'L': Modifiers.lower,
                  't': Modifiers.title}

        for k, v in modifiers.items():
            self.modifiers[k] = v

    def parse(self):
        result, tokens = self._parse(self.pattern, None)

        return result

    def _parse(self, tokens, terminator):
        result = ''

        while tokens != '':
            if self._peek(tokens) == '' or self._peek(tokens) == terminator:
                tokens = self._remains(tokens)
                break

            try:
                res, tokens = self._expr(tokens, terminator)
            except ParseError:
                res, tokens = self._text(tokens)

            result += res

        return result, tokens

    def _peek(self, tokens, num = 0):
        return (num < len(tokens) and tokens[num])

    def _token(self, tokens):
        if tokens == '':
            return '', '';

        return tokens[0], (len(tokens) > 1 and tokens[1:]) or ''

    def _remains(self, tokens, num = 1):
        return (num < len(tokens) and tokens[num:]) or ''

    def _expr(self, tokens, terminator):
        if tokens == '':
            return ''

        try:
            return {'\\': self._escape,
                '(': self._condition}[self._peek(tokens)](tokens, terminator)
        except KeyError:
            raise ParseError

    def _text(self, tokens):
        return self._token(tokens)

    def _substitute(self, group, modifiers = ''):
        result = (self.groups.has_key(group) and self.groups[group]) or ''

        for modifier in modifiers:
            if self.modifiers.has_key(modifier):
                result = self.modifiers[modifier](result)

        return result

    def _match_group(self, tokens):
        match = re.match('\\\\%s' % self.REG_GROUP, tokens)

        if not match:
            return None, tokens

        return self._substitute(match.group(1) or match.group(2), match.group(3) or ''), tokens[match.end():]

    def _escape(self, tokens, terminator):
        # Try to match a group
        result, tokens = self._match_group(tokens)

        if result != None:
            return result, tokens

        s = self.REG_GROUP

        if terminator:
            s += '|%s' % re.escape(terminator)

        match = re.match('\\\\(\\\\%s|%s)' % (s, self.REG_ESCAPE), tokens)

        if not match:
            raise ParseError

        return match.group(1), tokens[match.end():]

    def _condition_value(self, tokens):
        match = re.match('\\\\?%s\s*' % self.REG_GROUP, tokens)

        if not match:
            return None, tokens

        groups = match.groups()
        name = groups[0] or groups[1]

        return self.groups.has_key(name) and self.groups[name] != None, tokens[match.end():]

    def _condition(self, tokens, terminator):
        # Match ? after (
        if self._peek(tokens, 1) != '?':
            raise ParseError

        # Remove initial (? token
        tokens = self._remains(tokens, 2)
        condition, tokens = self._condition_value(tokens)

        if condition == None or self._peek(tokens) != ',':
            raise ParseError

        truepart, tokens = self._parse(self._remains(tokens), ',')

        if truepart == None:
            raise ParseError

        falsepart, tokens = self._parse(tokens, ')')

        if falsepart == None:
            raise ParseError

        if condition:
            return truepart, tokens
        else:
            return falsepart, tokens

    def escape_substitution(substitution):
        return re.sub('(%s|%s)' % (self.REG_GROUP, self.REG_ESCAPE), '\\\\\\1', substitution)

    escapesubstitution = staticmethod(escape_substitution)
# ex:ts=4:et:



# -*- coding: utf-8 -*-
#    Pluma External Tools plugin
#    Copyright (C) 2006  Steve Frécinaux <code@istique.net>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re
import locale
import codecs
from gi.repository import GLib


class Singleton(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls, *args, **kwargs)
            cls._instance.__init_once__()

        return cls._instance


class ToolLibrary(Singleton):
    def __init_once__(self):
        self.locations = []

    def set_locations(self, datadir):
        self.locations = []

        for d in self.get_xdg_data_dirs():
            self.locations.append(os.path.join(d, 'pluma', 'plugins', 'externaltools', 'tools'))

        self.locations.append(datadir)

        # self.locations[0] is where we save the custom scripts
        toolsdir = os.path.join(GLib.get_user_config_dir(), 'pluma/tools')
        self.locations.insert(0, toolsdir)

        if not os.path.isdir(self.locations[0]):
            os.makedirs(self.locations[0])
            self.tree = ToolDirectory(self, '')
            self.import_old_xml_store()
        else:
            self.tree = ToolDirectory(self, '')

    # cf. http://standards.freedesktop.org/basedir-spec/latest/
    def get_xdg_data_dirs(self):
        dirs = os.getenv('XDG_DATA_DIRS')
        if dirs:
            dirs = dirs.split(os.pathsep)
        else:
            dirs = ('/usr/local/share', '/usr/share')
        return dirs

    # This function is meant to be ran only once, when the tools directory is
    # created. It imports eventual tools that have been saved in the old XML
    # storage file.
    def import_old_xml_store(self):
        import xml.etree.ElementTree as et
        filename = os.path.join(GLib.get_user_config_dir(), 'pluma/pluma-tools.xml')

        if not os.path.isfile(filename):
            return

        print("External tools: importing old tools into the new store...")

        xtree = et.parse(filename)
        xroot = xtree.getroot()

        for xtool in xroot:
            for i in self.tree.tools:
                if i.name == xtool.get('label'):
                    tool = i
                    break
            else:
                tool = Tool(self.tree)
                tool.name = xtool.get('label')
                tool.autoset_filename()
                self.tree.tools.append(tool)
            tool.comment = xtool.get('description')
            tool.shortcut = xtool.get('accelerator')
            tool.applicability = xtool.get('applicability')
            tool.output = xtool.get('output')
            tool.input = xtool.get('input')

            tool.save_with_script(xtool.text)

    def get_full_path(self, path, mode='r', system=True, local=True):
        assert (system or local)
        if path is None:
            return None
        if mode == 'r':
            if system and local:
                locations = self.locations
            elif local and not system:
                locations = [self.locations[0]]
            elif system and not local:
                locations = self.locations[1:]
            else:
                raise ValueError("system and local can't be both set to False")

            for i in locations:
                p = os.path.join(i, path)
                if os.path.lexists(p):
                    return p
            return None
        else:
            path = os.path.join(self.locations[0], path)
            dirname = os.path.dirname(path)
            if not os.path.isdir(dirname):
                os.mkdir(dirname)
            return path


class ToolDirectory(object):
    def __init__(self, parent, dirname):
        super(ToolDirectory, self).__init__()
        self.subdirs = list()
        self.tools = list()
        if isinstance(parent, ToolDirectory):
            self.parent = parent
            self.library = parent.library
        else:
            self.parent = None
            self.library = parent
        self.dirname = dirname
        self._load()

    def listdir(self):
        elements = dict()
        for l in self.library.locations:
            d = os.path.join(l, self.dirname)
            if not os.path.isdir(d):
                continue
            for i in os.listdir(d):
                elements[i] = None
        keys = sorted(elements.keys())
        return keys

    def _load(self):
        for p in self.listdir():
            path = os.path.join(self.dirname, p)
            full_path = self.library.get_full_path(path)
            if os.path.isdir(full_path):
                self.subdirs.append(ToolDirectory(self, p))
            elif os.path.isfile(full_path) and os.access(full_path, os.X_OK):
                self.tools.append(Tool(self, p))

    def get_path(self):
        if self.parent is None:
            return self.dirname
        else:
            return os.path.join(self.parent.get_path(), self.dirname)
    path = property(get_path)

    def get_name(self):
        return os.path.basename(self.dirname)
    name = property(get_name)

    def delete_tool(self, tool):
        # Only remove it if it lays in $HOME
        if tool in self.tools:
            path = tool.get_path()
            if path is not None:
                filename = os.path.join(self.library.locations[0], path)
                if os.path.isfile(filename):
                    os.unlink(filename)
            self.tools.remove(tool)
            return True
        else:
            return False

    def revert_tool(self, tool):
        # Only remove it if it lays in $HOME
        filename = os.path.join(self.library.locations[0], tool.get_path())
        if tool in self.tools and os.path.isfile(filename):
            os.unlink(filename)
            tool._load()
            return True
        else:
            return False


class Tool(object):
    RE_KEY = re.compile('^([a-zA-Z_][a-zA-Z0-9_.\-]*)(\[([a-zA-Z_@]+)\])?$')

    def __init__(self, parent, filename=None):
        super(Tool, self).__init__()
        self.parent = parent
        self.library = parent.library
        self.filename = filename
        self.changed = False
        self._properties = dict()
        self._transform = {
            'Languages': [self._to_list, self._from_list]
        }
        self._load()

    def _to_list(self, value):
        if value.strip() == '':
            return []
        else:
            return [x.strip() for x in value.split(',')]

    def _from_list(self, value):
        return ','.join(value)

    def _parse_value(self, key, value):
        if key in self._transform:
            return self._transform[key][0](value)
        else:
            return value

    def _load(self):
        if self.filename is None:
            return

        filename = self.library.get_full_path(self.get_path())
        if filename is None:
            return

        fp = codecs.open(filename, 'r', encoding='utf-8')
        in_block = False
        lang = locale.getlocale(locale.LC_MESSAGES)[0]

        for line in fp:
            if not in_block:
                in_block = line.startswith('# [Pluma Tool]')
                continue
            if line.startswith('##') or line.startswith('# #'):
                continue
            if not line.startswith('# '):
                break

            try:
                (key, value) = [i.strip() for i in line[2:].split('=', 1)]
                m = self.RE_KEY.match(key)
                if m.group(3) is None:
                    self._properties[m.group(1)] = self._parse_value(m.group(1), value)
                elif lang is not None and lang.startswith(m.group(3)):
                    self._properties[m.group(1)] = self._parse_value(m.group(1), value)
            except ValueError:
                break
        fp.close()
        self.changed = False

    def _set_property_if_changed(self, key, value):
        if value != self._properties.get(key):
            self._properties[key] = value

            self.changed = True

    def is_global(self):
        return self.library.get_full_path(self.get_path(), local=False) is not None

    def is_local(self):
        return self.library.get_full_path(self.get_path(), system=False) is not None

    def get_path(self):
        if self.filename is not None:
            return os.path.join(self.parent.get_path(), self.filename)
        else:
            return None
    path = property(get_path)

    # This command is the one that is meant to be ran
    # (later, could have an Exec key or something)
    def get_command(self):
        return self.library.get_full_path(self.get_path())
    command = property(get_command)

    def get_applicability(self):
        applicability = self._properties.get('Applicability')
        if applicability:
            return applicability
        return 'all'

    def set_applicability(self, value):
        self._set_property_if_changed('Applicability', value)

    applicability = property(get_applicability, set_applicability)

    def get_name(self):
        name = self._properties.get('Name')
        if name:
            return name
        return os.path.basename(self.filename)

    def set_name(self, value):
        self._set_property_if_changed('Name', value)

    name = property(get_name, set_name)

    def get_shortcut(self):
        shortcut = self._properties.get('Shortcut')
        if shortcut:
            return shortcut
        return None

    def set_shortcut(self, value):
        self._set_property_if_changed('Shortcut', value)

    shortcut = property(get_shortcut, set_shortcut)

    def get_comment(self):
        comment = self._properties.get('Comment')
        if comment:
            return comment
        return self.filename

    def set_comment(self, value):
        self._set_property_if_changed('Comment', value)

    comment = property(get_comment, set_comment)

    def get_input(self):
        input = self._properties.get('Input')
        if input:
            return input
        return 'nothing'

    def set_input(self, value):
        self._set_property_if_changed('Input', value)

    input = property(get_input, set_input)

    def get_output(self):
        output = self._properties.get('Output')
        if output:
            return output
        return 'output-panel'

    def set_output(self, value):
        self._set_property_if_changed('Output', value)

    output = property(get_output, set_output)

    def get_save_files(self):
        save_files = self._properties.get('Save-files')
        if save_files:
            return save_files
        return 'nothing'

    def set_save_files(self, value):
        self._set_property_if_changed('Save-files', value)

    save_files = property(get_save_files, set_save_files)

    def get_languages(self):
        languages = self._properties.get('Languages')
        if languages:
            return languages
        return []

    def set_languages(self, value):
        self._set_property_if_changed('Languages', value)

    languages = property(get_languages, set_languages)

    def has_hash_bang(self):
        if self.filename is None:
            return True

        filename = self.library.get_full_path(self.get_path())
        if filename is None:
            return True

        fp = codecs.open(filename, 'r', encoding='utf-8')
        for line in fp:
            if line.strip() == '':
                continue
            return line.startswith('#!')

    # There is no property for this one because this function is quite
    # expensive to perform
    def get_script(self):
        if self.filename is None:
            return ["#!/bin/sh\n"]

        filename = self.library.get_full_path(self.get_path())
        if filename is None:
            return ["#!/bin/sh\n"]

        fp = codecs.open(filename, 'r', encoding='utf-8')
        lines = list()

        # before entering the data block
        for line in fp:
            if line.startswith('# [Pluma Tool]'):
                break
            lines.append(line)
        # in the block:
        for line in fp:
            if line.startswith('##'):
                continue
            if not (line.startswith('# ') and '=' in line):
                # after the block: strip one emtpy line (if present)
                if line.strip() != '':
                    lines.append(line)
                break
        # after the block
        for line in fp:
            lines.append(line)
        fp.close()
        return lines

    def _dump_properties(self):
        lines = ['# [Pluma Tool]']
        for item in self._properties.items():
            if item[0] in self._transform:
                lines.append('# %s=%s' % (item[0], self._transform[item[0]][1](item[1])))
            elif item[1] is not None:
                lines.append('# %s=%s' % item)
        return '\n'.join(lines) + '\n'

    def save_with_script(self, script):
        filename = self.library.get_full_path(self.filename, 'w')
        fp = codecs.open(filename, 'w', encoding='utf-8')

        # Make sure to first print header (shebang, modeline), then
        # properties, and then actual content
        header = []
        content = []
        inheader = True

        # Parse
        for line in script:
            line = line.rstrip("\n")
            if not inheader:
                content.append(line)
            elif line.startswith('#!'):
                # Shebang (should be always present)
                header.append(line)
            elif line.strip().startswith('#') and ('-*-' in line or 'ex:' in line or 'vi:' in line or 'vim:' in line):
                header.append(line)
            else:
                content.append(line)
                inheader = False

        # Write out header
        for line in header:
            fp.write(line + "\n")

        fp.write(self._dump_properties())
        fp.write("\n")

        for line in content:
            fp.write(line + "\n")

        fp.close()
        os.chmod(filename, 0o750)
        self.changed = False

    def save(self):
        if self.changed:
            self.save_with_script(self.get_script())

    def autoset_filename(self):
        if self.filename is not None:
            return
        dirname = self.parent.path
        if dirname != '':
            dirname += os.path.sep

        basename = self.name.lower().replace(' ', '-').replace('/', '-')

        if self.library.get_full_path(dirname + basename):
            i = 2
            while self.library.get_full_path(dirname + "%s-%d" % (basename, i)):
                i += 1
            basename = "%s-%d" % (basename, i)
        self.filename = basename

if __name__ == '__main__':
    library = ToolLibrary()
    library.set_locations(os.path.expanduser("~/.config/pluma/tools"))

    def print_tool(t, indent):
        print(indent * "  " + "%s: %s" % (t.filename, t.name))

    def print_dir(d, indent):
        print(indent * "  " + d.dirname + '/')
        for i in d.subdirs:
            print_dir(i, indent + 1)
        for i in d.tools:
            print_tool(i, indent + 1)

    print_dir(library.tree, 0)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#  Copyright (C) 2012-2021 MATE Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os
import sys
import fnmatch
import xml.sax.saxutils
from gi.repository import GObject, Gio, GLib, Gdk, Gtk, Pango, Pluma
from .virtualdirs import VirtualDirectory

class Popup(Gtk.Dialog):
    __gtype_name__ = "QuickOpenPopup"

    def __init__(self, window, paths, handler):
        Gtk.Dialog.__init__(self,
                            title=_('Quick Open'),
                            parent=window,
                            flags=Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL)

        self._add_button(_("_Cancel"), Gtk.ResponseType.CANCEL, "process-stop")
        self._open_button = self._add_button(_("_Open"),
                                             Gtk.ResponseType.ACCEPT,
                                             "document-open")
        self._handler = handler
        self._build_ui()

        self._size = (0, 0)
        self._dirs = []
        self._cache = {}
        self._theme = None
        self._cursor = None
        self._shift_start = None

        self._busy_cursor = Gdk.Cursor(Gdk.CursorType.WATCH)

        accel_group = Gtk.AccelGroup()
        accel_group.connect(Gdk.keyval_from_name('l'),
                            Gdk.ModifierType.CONTROL_MASK,
                            0,
                            self.on_focus_entry)

        self.add_accel_group(accel_group)

        unique = []

        for path in paths:
            if not path.get_uri() in unique:
                self._dirs.append(path)
                unique.append(path.get_uri())

    def get_final_size(self):
        return self._size

    def _build_ui(self):
        self.set_border_width(5)
        vbox = self.get_content_area()
        vbox.set_spacing(3)

        self._entry = Gtk.SearchEntry()

        self._entry.connect('changed', self.on_changed)
        self._entry.connect('key-press-event', self.on_key_press_event)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.OUT)

        tv = Gtk.TreeView()
        tv.set_headers_visible(False)

        self._store = Gtk.ListStore(Gio.Icon,
                                    str,
                                    GObject.Object,
                                    Gio.FileType)
        tv.set_model(self._store)

        self._treeview = tv
        tv.connect('row-activated', self.on_row_activated)

        renderer = Gtk.CellRendererPixbuf()
        column = Gtk.TreeViewColumn()
        column.pack_start(renderer, False)
        column.add_attribute(renderer, "gicon", 0)

        renderer = Gtk.CellRendererText()
        column.pack_start(renderer, True)
        column.add_attribute(renderer, "markup", 1)

        column.set_cell_data_func(renderer, self.on_cell_data_cb, None)

        tv.append_column(column)
        sw.add(tv)

        selection = tv.get_selection()
        selection.connect('changed', self.on_selection_changed)
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)

        vbox.pack_start(self._entry, False, False, 0)
        vbox.pack_start(sw, True, True, 0)

        lbl = Gtk.Label()
        lbl.set_halign(Gtk.Align.START)
        lbl.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
        self._info_label = lbl

        vbox.pack_start(lbl, False, False, 0)

        # Initial selection
        self.on_selection_changed(tv.get_selection())
        vbox.show_all()

    def on_cell_data_cb(self, column, cell, model, piter, user_data):
        path = model.get_path(piter)

        if self._cursor and path == self._cursor.get_path():
            style = self._treeview.get_style()
            bg = style.bg[Gtk.StateType.PRELIGHT]

            cell.set_property('cell-background-gdk', bg)
            cell.set_property('style', Pango.Style.ITALIC)
        else:
            cell.set_property('cell-background-set', False)
            cell.set_property('style-set', False)

    def _add_button(self, label, response, icon=None):
        button = self.add_button(label, response)
        if icon:
            image = Gtk.Image.new_from_icon_name(icon, Gtk.IconSize.BUTTON)
            button.set_image(image)
            button.set_property("always-show-image", True)
        return button

    def _list_dir(self, gfile):
        entries = []

        try:
            ret = gfile.enumerate_children("standard::*",
                                           Gio.FileQueryInfoFlags.NONE,
                                           None)
        except GLib.Error:
            pass

        if isinstance(ret, Gio.FileEnumerator):
            while True:
                entry = ret.next_file(None)

                if not entry:
                    break

                if not entry.get_is_backup():
                    entries.append((gfile.get_child(entry.get_name()), entry))
        else:
            entries = ret

        children = []

        for entry in entries:
            children.append((entry[0],
                             entry[1].get_name(),
                             entry[1].get_file_type(),
                             entry[1].get_icon()))

        return children

    def _key_entries(self, pos):
        return pos if pos >= 0 else sys.maxsize

    def _match_glob(self, s, glob):
        if glob:
            glob += '*'

        return fnmatch.fnmatch(s, glob)

    def do_search_dir(self, parts, d):
        if not parts or not d:
            return []

        if not d in self._cache:
            entries = self._list_dir(d)
            entries.sort(key=lambda x: x[1].lower())
            self._cache[d] = entries
        else:
            entries = self._cache[d]

        found = []
        newdirs = []

        lpart = parts[0].lower()

        for entry in entries:
            if not entry:
                continue

            lentry = entry[1].lower()

            if not lpart or lpart in lentry or self._match_glob(lentry, lpart):
                if entry[2] == Gio.FileType.DIRECTORY:
                    if len(parts) > 1:
                        newdirs.append(entry[0])
                    else:
                        found.append(entry)
                elif entry[2] == Gio.FileType.REGULAR and \
                        (not lpart or len(parts) == 1):
                    found.append(entry)

        found.sort(key=lambda x: self._key_entries(x[1].lower().find(lpart)))

        if lpart == '..':
            newdirs.append(d.get_parent())

        for dd in newdirs:
            found.extend(self.do_search_dir(parts[1:], dd))

        return found

    def _replace_insensitive(self, s, find, rep):
        out = ''
        l = s.lower()
        find = find.lower()
        last = 0

        if len(find) == 0:
            return xml.sax.saxutils.escape(s)

        while True:
            m = l.find(find, last)

            if m == -1:
                break
            else:
                out += xml.sax.saxutils.escape(s[last:m]) + rep % (xml.sax.saxutils.escape(s[m:m + len(find)]),)
                last = m + len(find)

        return out + xml.sax.saxutils.escape(s[last:])

    def make_markup(self, parts, path):
        out = []

        for i in range(0, len(parts)):
            out.append(self._replace_insensitive(path[i], parts[i], "<b>%s</b>"))

        return os.sep.join(out)

    def _get_icon(self, f):
        query = f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_ICON,
                             Gio.FileQueryInfoFlags.NONE,
                             None)

        if not query:
            return None
        else:
            return query.get_icon()

    def _make_parts(self, parent, child, pp):
        parts = []

        # We went from parent, to child, using pp
        idx = len(pp) - 1

        while idx >= 0:
            if pp[idx] == '..':
                parts.insert(0, '..')
            else:
                parts.insert(0, child.get_basename())
                child = child.get_parent()

            idx -= 1

        return parts

    def normalize_relative(self, parts):
        if not parts:
            return []

        out = self.normalize_relative(parts[:-1])

        if parts[-1] == '..':
            if not out or (out[-1] == '..') or len(out) == 1:
                out.append('..')
            else:
                del out[-1]
        else:
            out.append(parts[-1])

        return out

    def _append_to_store(self, item):
        if not item in self._stored_items:
            self._store.append(item)
            self._stored_items[item] = True

    def _clear_store(self):
        self._store.clear()
        self._stored_items = {}

    def _show_virtuals(self):
        for d in self._dirs:
            if isinstance(d, VirtualDirectory):
                for entry in d.enumerate_children("standard::*", 0, None):
                    self._append_to_store((entry[1].get_icon(),
                                           xml.sax.saxutils.escape(entry[1].get_name()),
                                           entry[0],
                                           entry[1].get_file_type()))

    def _set_busy(self, busy):
        if busy:
            self.get_window().set_cursor(self._busy_cursor)
        else:
            self.get_window().set_cursor(None)

        Gdk.flush()

    def _remove_cursor(self):
        if self._cursor:
            path = self._cursor.get_path()
            self._cursor = None

            self._store.row_changed(path, self._store.get_iter(path))

    def do_search(self):
        self._set_busy(True)
        self._remove_cursor()

        text = self._entry.get_text().strip()
        self._clear_store()

        if text == '':
            self._show_virtuals()
        else:
            parts = self.normalize_relative(text.split(os.sep))
            files = []

            for d in self._dirs:
                for entry in self.do_search_dir(parts, d):
                    pathparts = self._make_parts(d, entry[0], parts)
                    self._append_to_store((entry[3],
                                           self.make_markup(parts, pathparts),
                                           entry[0],
                                           entry[2]))

        piter = self._store.get_iter_first()

        if piter:
            path = self._store.get_path(piter)
            self._treeview.get_selection().select_path(path)

        self._set_busy(False)

    def do_show(self):
        Gtk.Window.do_show(self)

        self._entry.grab_focus()
        self._entry.set_text("")

        self.do_search()

    def on_changed(self, editable):
        self.do_search()
        self.on_selection_changed(self._treeview.get_selection())

    def _shift_extend(self, towhere):
        selection = self._treeview.get_selection()

        if not self._shift_start:
            model, rows = selection.get_selected_rows()
            start = rows[0]

            self._shift_start = Gtk.TreeRowReference.new(self._store, start)
        else:
            start = self._shift_start.get_path()

        selection.unselect_all()
        selection.select_range(start, towhere)

    def _select_index(self, idx, hasctrl, hasshift):
        path = (idx,)

        if not (hasctrl or hasshift):
            self._treeview.get_selection().unselect_all()

        if hasshift:
            self._shift_extend(path)
        else:
            self._shift_start = None

            if not hasctrl:
                self._treeview.get_selection().select_path(path)

        self._treeview.scroll_to_cell(path, None, True, 0.5, 0)
        self._remove_cursor()

        if hasctrl or hasshift:
            self._cursor = Gtk.TreeRowReference(self._store, path)

            piter = self._store.get_iter(path)
            self._store.row_changed(path, piter)

    def _move_selection(self, howmany, hasctrl, hasshift):
        num = self._store.iter_n_children(None)

        if num == 0:
            return True

        # Test for cursor
        path = None

        if self._cursor:
            path = self._cursor.get_path()
        else:
            model, rows = self._treeview.get_selection().get_selected_rows()

            if len(rows) == 1:
                path = rows[0]

        if not path:
            if howmany > 0:
                self._select_index(0, hasctrl, hasshift)
            else:
                self._select_index(num - 1, hasctrl, hasshift)
        else:
            idx = path.get_indices()[0]

            if idx + howmany < 0:
                self._select_index(0, hasctrl, hasshift)
            elif idx + howmany >= num:
                self._select_index(num - 1, hasctrl, hasshift)
            else:
                self._select_index(idx + howmany, hasctrl, hasshift)

        return True

    def _direct_file(self):
        uri = self._entry.get_text()
        gfile = None

        if Pluma.utils_is_valid_uri(uri):
            gfile = Gio.file_new_for_uri(uri)
        elif os.path.isabs(uri):
            f = Gio.file_new_for_uri(uri)

            if f.query_exists():
                gfile = f

        return gfile

    def _activate(self):
        model, rows = self._treeview.get_selection().get_selected_rows()
        ret = True

        for row in rows:
            s = model.get_iter(row)
            info = model.get(s, 2, 3)

            if info[1] != Gio.FileType.DIRECTORY:
                ret = ret and self._handler(info[0])
            else:
                text = self._entry.get_text()

                for i in range(len(text) - 1, -1, -1):
                    if text[i] == os.sep:
                        break

                self._entry.set_text(os.path.join(text[:i],
                                  os.path.basename(info[0].get_uri())) + os.sep)
                self._entry.set_position(-1)
                self._entry.grab_focus()
                return True

        if rows and ret:
            # We destroy the popup in an idle callback to work around a crash that happens with
            # GTK_IM_MODULE=xim.  See https://bugzilla.gnome.org/show_bug.cgi?id=737711 .
            GLib.idle_add(self.destroy)

        if not rows:
            gfile = self._direct_file()

            if gfile and self._handler(gfile):
                GLib.idle_add(self.destroy)
            else:
                ret = False
        else:
            ret = False

        return ret

    def toggle_cursor(self):
        if not self._cursor:
            return

        path = self._cursor.get_path()
        selection = self._treeview.get_selection()

        if selection.path_is_selected(path):
            selection.unselect_path(path)
        else:
            selection.select_path(path)

    def on_key_press_event(self, widget, event):
        move_mapping = {
            "Down": 1,
            "Up": -1,
            "Page_Down": 5,
            "Page_Up": -5
        }

        keyname = Gdk.keyval_name(event.keyval)

        if keyname == "Escape":
            self.destroy()
            return True
        elif keyname in move_mapping:
            return self._move_selection(move_mapping[keyname],
                                   event.state & Gdk.ModifierType.CONTROL_MASK,
                                   event.state & Gdk.ModifierType.SHIFT_MASK)
        elif keyname in ["Return", "KP_Enter", "Tab", "ISO_Left_Tab"]:
            return self._activate()
        elif keyname == "space" and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.toggle_cursor()

        return False

    def on_row_activated(self, view, path, column):
        self._activate()

    def do_response(self, response):
        if response != Gtk.ResponseType.ACCEPT or not self._activate():
            self.destroy()

    def do_configure_event(self, event):
        if self.get_realized():
            alloc = self.get_allocation()
            self._size = (alloc.width, alloc.height)

        return Gtk.Dialog.do_configure_event(self, event)

    def on_selection_changed(self, selection):
        model, rows = selection.get_selected_rows()

        gfile = None
        fname = None

        if not rows:
            gfile = self._direct_file()
        elif len(rows) == 1:
            gfile = model.get(model.get_iter(rows[0]), 2)[0]
        else:
            fname = ''

        if gfile:
            if gfile.is_native():
                fname = xml.sax.saxutils.escape(gfile.get_path())
            else:
                fname = xml.sax.saxutils.escape(gfile.get_uri())

        self._open_button.set_sensitive(fname != None)
        self._info_label.set_markup(fname or '')

    def on_focus_entry(self, group, accel, keyval, modifier):
        self._entry.grab_focus()

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# config.py -- Config dialog
#
# Copyright (C) 2008 - B. Clausius
# Copyright (C) 2012-2021 MATE Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

# Parts from "Interactive Python-GTK Console" (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from pluma Python Console Plugin
#     Copyrignt (C), 2005 Raphaël Slinckx

import os
from gi.repository import Gio, Gtk, Gdk

__all__ = ('PythonConsoleConfig', 'PythonConsoleConfigWidget')

class PythonConsoleConfig(object):

    CONSOLE_KEY_BASE = 'org.mate.pluma.plugins.pythonconsole'
    CONSOLE_KEY_COMMAND_COLOR = 'command-color'
    CONSOLE_KEY_ERROR_COLOR = 'error-color'
    CONSOLE_KEY_USE_SYSTEM_FONT = 'use-system-font'
    CONSOLE_KEY_FONT = 'font'

    INTERFACE_KEY_BASE = 'org.mate.interface'
    INTERFACE_KEY_MONOSPACE_FONT_NAME = 'monospace-font-name'

    color_command = property(
        lambda self: self.console_settings.get_string(self.CONSOLE_KEY_COMMAND_COLOR),
        lambda self, value: self.console_settings.set_string(self.CONSOLE_KEY_COMMAND_COLOR, value)
    )

    color_error = property(
        lambda self: self.console_settings.get_string(self.CONSOLE_KEY_ERROR_COLOR),
        lambda self, value: self.console_settings.set_string(self.CONSOLE_KEY_ERROR_COLOR, value)
    )

    use_system_font = property(
        lambda self: self.console_settings.get_boolean(self.CONSOLE_KEY_USE_SYSTEM_FONT),
        lambda self, value: self.console_settings.set_boolean(self.CONSOLE_KEY_USE_SYSTEM_FONT, value)
    )

    font = property(
        lambda self: self.console_settings.get_string(self.CONSOLE_KEY_FONT),
        lambda self, value: self.console_settings.set_string(self.CONSOLE_KEY_FONT, value)
    )

    monospace_font_name = property(
        lambda self: self.interface_settings.get_string(self.INTERFACE_KEY_MONOSPACE_FONT_NAME)
    )

    console_settings = Gio.Settings.new(CONSOLE_KEY_BASE)
    interface_settings = Gio.Settings.new(INTERFACE_KEY_BASE)

    def __init__(self):
        object.__init__(self)

    @classmethod
    def enabled(self):
        return self.console_settings != None

    @classmethod
    def add_handler(self, handler):
        self.console_settings.connect("changed", handler)
        self.interface_settings.connect("changed", handler)


class PythonConsoleConfigWidget(object):

    CONSOLE_KEY_BASE = 'org.mate.pluma.plugins.pythonconsole'
    CONSOLE_KEY_COMMAND_COLOR = 'command-color'
    CONSOLE_KEY_ERROR_COLOR = 'error-color'

    def __init__(self, datadir):
        object.__init__(self)
        self._widget = None
        self._ui_path = os.path.join(datadir, 'ui', 'config.ui')
        self._config = PythonConsoleConfig()
        self._ui = Gtk.Builder()

    def configure_widget(self):
        if self._widget is None:
            self._ui.add_from_file(self._ui_path)

            self.set_colorbutton_color(self._ui.get_object('colorbutton-command'),
                                        self._config.color_command)
            self.set_colorbutton_color(self._ui.get_object('colorbutton-error'),
                                        self._config.color_error)
            checkbox = self._ui.get_object('checkbox-system-font')
            checkbox.set_active(self._config.use_system_font)
            self._fontbutton = self._ui.get_object('fontbutton-font')
            self._fontbutton.set_font_name(self._config.font)
            self.on_checkbox_system_font_toggled(checkbox)
            self._ui.connect_signals(self)

            self._widget = self._ui.get_object('widget-config')
            self._widget.show_all()

        return self._widget

    @staticmethod
    def set_colorbutton_color(colorbutton, value):
        rgba = Gdk.RGBA()
        parsed = rgba.parse(value)

        if parsed:
            colorbutton.set_rgba(rgba)

    def on_colorbutton_command_color_set(self, colorbutton):
        self._config.color_command = colorbutton.get_color().to_string()

    def on_colorbutton_error_color_set(self, colorbutton):
        self._config.color_error = colorbutton.get_color().to_string()

    def on_checkbox_system_font_toggled(self, checkbox):
        val = checkbox.get_active()
        self._config.use_system_font = val
        self._fontbutton.set_sensitive(not val)

    def on_fontbutton_font_set(self, fontbutton):
        self._config.font = fontbutton.get_font_name()

    def on_widget_config_parent_set(self, widget, oldparent):
        # Set icon in dialog close button.
        try:
            actionarea = widget.get_toplevel().get_action_area()
            image = Gtk.Image.new_from_icon_name("window-close",
                                                 Gtk.IconSize.BUTTON)
            for button in actionarea.get_children():
                button.set_image(image)
                button.set_property("always-show-image", True)
        except:
            pass

# ex:et:ts=4:



# -*- coding: utf-8 -*-

# pythonconsole.py -- Console widget
#
# Copyright (C) 2006 - Steve Frécinaux
# Copyright (C) 2012-2021 MATE Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

# Parts from "Interactive Python-GTK Console" (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from pluma Python Console Plugin
#     Copyrignt (C), 2005 Raphaël Slinckx

import string
import sys
import re
import traceback
from gi.repository import GObject, Gdk, Gtk, Pango

from .config import PythonConsoleConfig

__all__ = ('PythonConsole', 'OutFile')

class PythonConsole(Gtk.ScrolledWindow):

    __gsignals__ = {
        'grab-focus' : 'override',
    }

    DEFAULT_FONT = "Monospace 10"

    def __init__(self, namespace = {}):
        Gtk.ScrolledWindow.__init__(self)

        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.set_shadow_type(Gtk.ShadowType.IN)
        self.view = Gtk.TextView()
        self.view.set_editable(True)
        self.view.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.add(self.view)
        self.view.show()

        buffer = self.view.get_buffer()
        self.normal = buffer.create_tag("normal")
        self.error  = buffer.create_tag("error")
        self.command = buffer.create_tag("command")

        self.config = PythonConsoleConfig()
        self.config.add_handler(self.apply_preferences)
        self.apply_preferences()

        self.__spaces_pattern = re.compile(r'^\s+')
        self.namespace = namespace

        self.block_command = False

        # Init first line
        buffer.create_mark("input-line", buffer.get_end_iter(), True)
        buffer.insert(buffer.get_end_iter(), ">>> ")
        buffer.create_mark("input", buffer.get_end_iter(), True)

        # Init history
        self.history = ['']
        self.history_pos = 0
        self.current_command = ''
        self.namespace['__history__'] = self.history

        # Set up hooks for standard output.
        self.stdout = OutFile(self, sys.stdout.fileno(), self.normal)
        self.stderr = OutFile(self, sys.stderr.fileno(), self.error)

        # Signals
        self.view.connect("key-press-event", self.__key_press_event_cb)
        buffer.connect("mark-set", self.__mark_set_cb)

    def do_grab_focus(self):
        self.view.grab_focus()

    def apply_preferences(self, *args):
        self.error.set_property("foreground", self.config.color_error)
        self.command.set_property("foreground", self.config.color_command)

        if self.config.use_system_font:
            font_name = self.config.monospace_font_name
        else:
            font_name = self.config.font

        font_desc = None
        try:
            font_desc = Pango.FontDescription(font_name)
        except:
            try:
                font_desc = Pango.FontDescription(self.config.monospace_font_name)
            except:
                try:
                    font_desc = Pango.FontDescription(self.DEFAULT_FONT)
                except:
                    pass

        if font_desc:
            self.view.modify_font(font_desc)

    def stop(self):
        self.namespace = None

    def __key_press_event_cb(self, view, event):
        modifier_mask = Gtk.accelerator_get_default_mod_mask()
        event_state = event.state & modifier_mask
        keyname = Gdk.keyval_name(event.keyval)

        if keyname == "d" and event_state == Gdk.ModifierType.CONTROL_MASK:
            self.destroy()

        elif keyname == "Return" and \
             event_state == Gdk.ModifierType.CONTROL_MASK:
            # Get the command
            buffer = view.get_buffer()
            inp_mark = buffer.get_mark("input")
            inp = buffer.get_iter_at_mark(inp_mark)
            cur = buffer.get_end_iter()
            line = buffer.get_text(inp, cur, False)
            self.current_command = self.current_command + line + "\n"
            self.history_add(line)

            # Prepare the new line
            cur = buffer.get_end_iter()
            buffer.insert(cur, "\n... ")
            cur = buffer.get_end_iter()
            buffer.move_mark(inp_mark, cur)

            # Keep indentation of precendent line
            spaces = re.match(self.__spaces_pattern, line)
            if spaces is not None:
                buffer.insert(cur, line[spaces.start() : spaces.end()])
                cur = buffer.get_end_iter()

            buffer.place_cursor(cur)
            GObject.idle_add(self.scroll_to_end)
            return True

        elif keyname == "Return":
            # Get the marks
            buffer = view.get_buffer()
            lin_mark = buffer.get_mark("input-line")
            inp_mark = buffer.get_mark("input")

            # Get the command line
            inp = buffer.get_iter_at_mark(inp_mark)
            cur = buffer.get_end_iter()
            line = buffer.get_text(inp, cur, False)
            self.current_command = self.current_command + line + "\n"
            self.history_add(line)

            # Make the line blue
            lin = buffer.get_iter_at_mark(lin_mark)
            buffer.apply_tag(self.command, lin, cur)
            buffer.insert(cur, "\n")

            cur_strip = self.current_command.rstrip()

            if cur_strip.endswith(":") \
            or (self.current_command[-2:] != "\n\n" and self.block_command):
                # Unfinished block command
                self.block_command = True
                com_mark = "... "
            elif cur_strip.endswith("\\"):
                com_mark = "... "
            else:
                # Eval the command
                self.__run(self.current_command)
                self.current_command = ''
                self.block_command = False
                com_mark = ">>> "

            # Prepare the new line
            cur = buffer.get_end_iter()
            buffer.move_mark(lin_mark, cur)
            buffer.insert(cur, com_mark)
            cur = buffer.get_end_iter()
            buffer.move_mark(inp_mark, cur)
            buffer.place_cursor(cur)
            GObject.idle_add(self.scroll_to_end)
            return True

        elif keyname == "KP_Down" or keyname == "Down":
            # Next entry from history
            view.emit_stop_by_name("key_press_event")
            self.history_down()
            GObject.idle_add(self.scroll_to_end)
            return True

        elif keyname == "KP_Up" or keyname == "Up":
            # Previous entry from history
            view.emit_stop_by_name("key_press_event")
            self.history_up()
            GObject.idle_add(self.scroll_to_end)
            return True

        elif keyname == "KP_Left" or keyname == "Left" or \
             keyname == "BackSpace":
            buffer = view.get_buffer()
            inp = buffer.get_iter_at_mark(buffer.get_mark("input"))
            cur = buffer.get_iter_at_mark(buffer.get_insert())
            if inp.compare(cur) == 0:
                if not event_state:
                    buffer.place_cursor(inp)
                return True
            return False

        # For the console we enable smart/home end behavior incoditionally
        # since it is useful when editing python

        elif (keyname == "KP_Home" or keyname == "Home") and \
             event_state == event_state & (Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK):
            # Go to the begin of the command instead of the begin of the line
            buffer = view.get_buffer()
            iter = buffer.get_iter_at_mark(buffer.get_mark("input"))
            ins = buffer.get_iter_at_mark(buffer.get_insert())

            while iter.get_char().isspace():
                iter.forward_char()

            if iter.equal(ins):
                iter = buffer.get_iter_at_mark(buffer.get_mark("input"))

            if event_state & Gdk.ModifierType.SHIFT_MASK:
                buffer.move_mark_by_name("insert", iter)
            else:
                buffer.place_cursor(iter)
            return True

        elif (keyname == "KP_End" or keyname == "End") and \
             event_state == event_state & (Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK):

            buffer = view.get_buffer()
            iter = buffer.get_end_iter()
            ins = buffer.get_iter_at_mark(buffer.get_insert())

            iter.backward_char()

            while iter.get_char().isspace():
                iter.backward_char()

            iter.forward_char()

            if iter.equal(ins):
                iter = buffer.get_end_iter()

            if event_state & Gdk.ModifierType.SHIFT_MASK:
                buffer.move_mark_by_name("insert", iter)
            else:
                buffer.place_cursor(iter)
            return True

    def __mark_set_cb(self, buffer, iter, name):
        input = buffer.get_iter_at_mark(buffer.get_mark("input"))
        pos   = buffer.get_iter_at_mark(buffer.get_insert())
        self.view.set_editable(pos.compare(input) != -1)

    def get_command_line(self):
        buffer = self.view.get_buffer()
        inp = buffer.get_iter_at_mark(buffer.get_mark("input"))
        cur = buffer.get_end_iter()
        return buffer.get_text(inp, cur, False)

    def set_command_line(self, command):
        buffer = self.view.get_buffer()
        mark = buffer.get_mark("input")
        inp = buffer.get_iter_at_mark(mark)
        cur = buffer.get_end_iter()
        buffer.delete(inp, cur)
        buffer.insert(inp, command)
        self.view.grab_focus()

    def history_add(self, line):
        if line.strip() != '':
            self.history_pos = len(self.history)
            self.history[self.history_pos - 1] = line
            self.history.append('')

    def history_up(self):
        if self.history_pos > 0:
            self.history[self.history_pos] = self.get_command_line()
            self.history_pos = self.history_pos - 1
            self.set_command_line(self.history[self.history_pos])

    def history_down(self):
        if self.history_pos < len(self.history) - 1:
            self.history[self.history_pos] = self.get_command_line()
            self.history_pos = self.history_pos + 1
            self.set_command_line(self.history[self.history_pos])

    def scroll_to_end(self):
        iter = self.view.get_buffer().get_end_iter()
        self.view.scroll_to_iter(iter, 0.0, False, 0.5, 0.5)
        return False

    def write(self, text, tag = None):
        buffer = self.view.get_buffer()
        if tag is None:
            buffer.insert(buffer.get_end_iter(), text)
        else:
            buffer.insert_with_tags(buffer.get_end_iter(), text, tag)
        GObject.idle_add(self.scroll_to_end)

    def eval(self, command, display_command = False):
        buffer = self.view.get_buffer()
        lin = buffer.get_mark("input-line")
        buffer.delete(buffer.get_iter_at_mark(lin),
                      buffer.get_end_iter())

        if isinstance(command, list) or isinstance(command, tuple):
            for c in command:
                if display_command:
                    self.write(">>> " + c + "\n", self.command)
                self.__run(c)
        else:
            if display_command:
                self.write(">>> " + c + "\n", self.command)
            self.__run(command)

        cur = buffer.get_end_iter()
        buffer.move_mark_by_name("input-line", cur)
        buffer.insert(cur, ">>> ")
        cur = buffer.get_end_iter()
        buffer.move_mark_by_name("input", cur)
        self.view.scroll_to_iter(buffer.get_end_iter(), 0.0, False, 0.5, 0.5)

    def __run(self, command):
        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr

        # eval and exec are broken in how they deal with utf8-encoded
        # strings so we have to explicitly decode the command before
        # passing it along
        try:
            command = command.decode('utf8')
        except:
            pass

        try:
            try:
                r = eval(command, self.namespace, self.namespace)
                if r is not None:
                    print(repr(r))
            except SyntaxError:
                exec(command, self.namespace)
        except:
            if hasattr(sys, 'last_type') and sys.last_type == SystemExit:
                self.destroy()
            else:
                traceback.print_exc()

        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr

    def destroy(self):
        pass
        #Gtk.ScrolledWindow.destroy(self)

class OutFile:
    """A fake output file object. It sends output to a TK test widget,
    and if asked for a file number, returns one set on instance creation"""
    def __init__(self, console, fn, tag):
        self.fn = fn
        self.console = console
        self.tag = tag
    def close(self):         pass
    def flush(self):         pass
    def fileno(self):        return self.fn
    def isatty(self):        return 0
    def read(self, a):       return ''
    def readline(self):      return ''
    def readlines(self):     return []
    def write(self, s):      self.console.write(s, self.tag)
    def writelines(self, l): self.console.write(l, self.tag)
    def seek(self, a):       raise IOError(29, 'Illegal seek')
    def tell(self):          raise IOError(29, 'Illegal seek')
    truncate = tell

# ex:et:ts=4:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re

from gi.repository import GLib, Gio, Gdk, Gtk, Pluma

from .Library import Library
from .Snippet import Snippet
from .Placeholder import *
from . import Completion

class DynamicSnippet(dict):
    def __init__(self, text):
        self['text'] = text
        self.valid = True

class Document:
    TAB_KEY_VAL = ('Tab', 'ISO_Left_Tab')

    def __init__(self, instance, view):
        self.view = None
        self.instance = instance

        self.placeholders = []
        self.active_snippets = []
        self.active_placeholder = None
        self.signal_ids = {}

        self.ordered_placeholders = []
        self.update_placeholders = []
        self.jump_placeholders = []
        self.language_id = 0
        self.timeout_update_id = 0

        self.provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)

        # Always have a reference to the global snippets
        Library().ref(None)
        self.set_view(view)

    # Stop controlling the view. Remove all active snippets, remove references
    # to the view and the plugin instance, disconnect all signal handlers
    def stop(self):
        if self.timeout_update_id != 0:
            GLib.source_remove(self.timeout_update_id)
            self.timeout_update_id = 0
            del self.update_placeholders[:]
            del self.jump_placeholders[:]

        # Always release the reference to the global snippets
        Library().unref(None)
        self.set_view(None)
        self.instance = None
        self.active_placeholder = None

    def disconnect_signal(self, obj, signal):
        if (obj, signal) in self.signal_ids:
            obj.disconnect(self.signal_ids[(obj, signal)])
            del self.signal_ids[(obj, signal)]

    def connect_signal(self, obj, signal, cb):
        self.disconnect_signal(obj, signal)
        self.signal_ids[(obj, signal)] = obj.connect(signal, cb)

    def connect_signal_after(self, obj, signal, cb):
        self.disconnect_signal(obj, signal)
        self.signal_ids[(obj, signal)] = obj.connect_after(signal, cb)

    # Set the view to be controlled. Installs signal handlers and sets current
    # language. If there is already a view set this function will first remove
    # all currently active snippets and disconnect all current signals. So
    # self.set_view(None) will effectively remove all the control from the
    # current view
    def _set_view(self, view):
        if self.view:
            buf = self.view.get_buffer()

            # Remove signals
            signals = {self.view: ('key-press-event', 'destroy',
                           'notify::editable', 'drag-data-received', 'expose-event'),
                   buf:       ('notify::language', 'changed', 'cursor-moved', 'insert-text'),
                   self.view.get_completion(): ('hide',)}

            for obj, sig in signals.items():
                if obj:
                    for s in sig:
                        self.disconnect_signal(obj, s)

            # Remove all active snippets
            for snippet in list(self.active_snippets):
                self.deactivate_snippet(snippet, True)

            completion = self.view.get_completion()
            if completion and self.provider in completion.get_providers():
                completion.remove_provider(self.provider)

        self.view = view

        if view != None:
            buf = view.get_buffer()

            self.connect_signal(view, 'destroy', self.on_view_destroy)

            if view.get_editable():
                self.connect_signal(view, 'key-press-event', self.on_view_key_press)

            self.connect_signal(buf, 'notify::language', self.on_notify_language)
            self.connect_signal(view, 'notify::editable', self.on_notify_editable)
            self.connect_signal(view, 'drag-data-received', self.on_drag_data_received)
            self.connect_signal_after(view, 'draw', self.on_draw)

            self.update_language()

            completion = view.get_completion()
            completion.add_provider(self.provider)
        elif self.language_id != 0:
            langid = self.language_id

            self.language_id = None;
            self.provider.language_id = self.language_id

            if self.instance:
                self.instance.language_changed(self)

            Library().unref(langid)

    def set_view(self, view):
        if view == self.view:
            return

        self._set_view(view)

    # Call this whenever the language in the view changes. This makes sure that
    # the correct language is used when finding snippets
    def update_language(self):
        lang = self.view.get_buffer().get_language()

        if lang == None and self.language_id == None:
            return
        elif lang and lang.get_id() == self.language_id:
            return

        langid = self.language_id

        if lang:
            self.language_id = lang.get_id()
        else:
            self.language_id = None

        if self.instance:
            self.instance.language_changed(self)

        if langid != 0:
            Library().unref(langid)

        Library().ref(self.language_id)
        self.provider.language_id = self.language_id

    def accelerator_activate(self, keyval, mod):
        if not self.view or not self.view.get_editable():
            return False

        accelerator = Gtk.accelerator_name(keyval, mod)
        snippets = Library().from_accelerator(accelerator, \
                self.language_id)

        if len(snippets) == 0:
            return False
        elif len(snippets) == 1:
            self.apply_snippet(snippets[0])
        else:
            # Do the fancy completion dialog
            provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
            provider.set_proposals(snippets)

            cm = self.view.get_completion()
            cm.show([provider], cm.create_context(None))

        return True

    def first_snippet_inserted(self):
        buf = self.view.get_buffer()

        self.connect_signal(buf, 'changed', self.on_buffer_changed)
        self.connect_signal(buf, 'cursor-moved', self.on_buffer_cursor_moved)
        self.connect_signal_after(buf, 'insert-text', self.on_buffer_insert_text)

    def last_snippet_removed(self):
        buf = self.view.get_buffer()
        self.disconnect_signal(buf, 'changed')
        self.disconnect_signal(buf, 'cursor-moved')
        self.disconnect_signal(buf, 'insert-text')

    def current_placeholder(self):
        buf = self.view.get_buffer()

        piter = buf.get_iter_at_mark(buf.get_insert())
        found = []

        for placeholder in self.placeholders:
            begin = placeholder.begin_iter()
            end = placeholder.end_iter()

            if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                found.append(placeholder)

        if self.active_placeholder in found:
            return self.active_placeholder
        elif len(found) > 0:
            return found[0]
        else:
            return None

    def advance_placeholder(self, direction):
        # Returns (CurrentPlaceholder, NextPlaceholder), depending on direction
        buf = self.view.get_buffer()

        piter = buf.get_iter_at_mark(buf.get_insert())
        found = current = next = None
        length = len(self.placeholders)

        placeholders = list(self.placeholders)

        if self.active_placeholder:
            begin = self.active_placeholder.begin_iter()
            end = self.active_placeholder.end_iter()

            if piter.compare(begin) >= 0 and piter.compare(end) <= 0:
                current = self.active_placeholder
                currentIndex = placeholders.index(self.active_placeholder)

        if direction == 1:
            # w = piter, x = begin, y = end, z = found
            nearest = lambda w, x, y, z: (w.compare(x) <= 0 and (not z or \
                    x.compare(z.begin_iter()) < 0))
            indexer = lambda x: x < length - 1
        else:
            # w = piter, x = begin, y = end, z = prev
            nearest = lambda w, x, y, z: (w.compare(x) >= 0 and (not z or \
                    x.compare(z.begin_iter()) >= 0))
            indexer = lambda x: x > 0

        for index in range(0, length):
            placeholder = placeholders[index]
            begin = placeholder.begin_iter()
            end = placeholder.end_iter()

            # Find the nearest placeholder
            if nearest(piter, begin, end, found):
                found = placeholder

            # Find the current placeholder
            if piter.compare(begin) >= 0 and \
                    piter.compare(end) <= 0 and \
                    current == None:
                currentIndex = index
                current = placeholder

        if current and found and current != found and \
           (current.begin_iter().compare(found.begin_iter()) == 0 or \
            current.end_iter().compare(found.begin_iter()) == 0) and \
           self.active_placeholder and \
           current.begin_iter().compare(self.active_placeholder.begin_iter()) == 0:
            # if current and found are at the same place, then
            # resolve the 'hugging' problem
            current = self.active_placeholder
            currentIndex = placeholders.index(current)

        if current:
            if indexer(currentIndex):
                next = placeholders[currentIndex + direction]
        elif found:
            next = found
        elif length > 0:
            next = self.placeholders[0]

        return current, next

    def next_placeholder(self):
        return self.advance_placeholder(1)

    def previous_placeholder(self):
        return self.advance_placeholder(-1)

    def cursor_on_screen(self):
        buf = self.view.get_buffer()
        self.view.scroll_mark_onscreen(buf.get_insert())

    def set_active_placeholder(self, placeholder):
        self.active_placeholder = placeholder

    def goto_placeholder(self, current, next):
        last = None

        if current:
            # Signal this placeholder to end action
            self.view.get_completion().hide()
            current.leave()

            if current.__class__ == PlaceholderEnd:
                last = current

        self.set_active_placeholder(next)

        if next:
            next.enter()

            if next.__class__ == PlaceholderEnd:
                last = next
            elif len(next.defaults) > 1 and next.get_text() == next.default:
                provider = Completion.Defaults(self.on_default_activated)
                provider.set_defaults(next.defaults)

                cm = self.view.get_completion()
                cm.show([provider], cm.create_context(None))

        if last:
            # This is the end of the placeholder, remove the snippet etc
            for snippet in list(self.active_snippets):
                if snippet.placeholders[0] == last:
                    self.deactivate_snippet(snippet)
                    break

        self.cursor_on_screen()

        return next != None

    def skip_to_next_placeholder(self):
        (current, next) = self.next_placeholder()
        return self.goto_placeholder(current, next)

    def skip_to_previous_placeholder(self):
        (current, prev) = self.previous_placeholder()
        return self.goto_placeholder(current, prev)

    def env_get_selected_text(self, buf):
        bounds = buf.get_selection_bounds()

        if bounds:
            return buf.get_text(bounds[0], bounds[1], False)
        else:
            return ''

    def env_get_current_word(self, buf):
        start, end = buffer_word_boundary(buf)
        return buf.get_text(start, end, False)

    def env_get_current_line(self, buf):
        start, end = buffer_line_boundary(buf)
        return buf.get_text(start, end, False)

    def env_get_current_line_number(self, buf):
        start, end = buffer_line_boundary(buf)
        return str(start.get_line() + 1)

    def location_uri_for_env(self, location):
        if location:
            return location.get_uri()
        return ''

    def location_name_for_env(self, location):
        if location:
            return location.get_basename()
        return ''

    def location_scheme_for_env(self, location):
        if location:
            return location.get_uri_scheme()
        return ''

    def location_path_for_env(self, location):
        if location:
            return location.get_path()
        return ''

    def location_dir_for_env(self, location):
        if location:
            parent = location.get_parent()

            if parent and parent.has_uri_scheme('file'):
                return parent.get_path() or ''

        return ''

    def env_add_for_location(self, environ, location, prefix):
        parts = {
            'URI': self.location_uri_for_env,
            'NAME': self.location_name_for_env,
            'SCHEME': self.location_scheme_for_env,
            'PATH': self.location_path_for_env,
            'DIR': self.location_dir_for_env,
        }

        for k in parts:
            v = parts[k](location)
            key = prefix + '_' + k
            environ[key] = str(v)

        return environ

    def env_get_document_type(self, buf):
        typ = buf.get_mime_type()

        if typ:
            return typ
        else:
            return ''

    def env_get_documents_uri(self, buf):
        toplevel = self.view.get_toplevel()

        if isinstance(toplevel, Pluma.Window):
            documents_uri = [doc.get_location().get_uri()
                     for doc in toplevel.get_documents()
                     if doc.get_location() is not None]
        else:
            documents_uri = []

        return ' '.join(documents_uri)

    def env_get_documents_path(self, buf):
        toplevel = self.view.get_toplevel()

        if isinstance(toplevel, Pluma.Window):
            documents_location = [doc.get_location()
                          for doc in toplevel.get_documents()
                          if doc.get_location() is not None]

            documents_path = [location.get_path()
                      for location in documents_location
                      if Pluma.utils_uri_has_file_scheme(location.get_uri())]
        else:
            documents_path = []

        return ' '.join(documents_path)

    def get_environment(self):
        buf = self.view.get_buffer()
        environ = {}

        for k in os.environ:
            # Get the original environment
            v = os.environ[k]
            environ[k] = v

        variables = {
                 'PLUMA_SELECTED_TEXT': self.env_get_selected_text,
                 'PLUMA_CURRENT_WORD': self.env_get_current_word,
                 'PLUMA_CURRENT_LINE': self.env_get_current_line,
                 'PLUMA_CURRENT_LINE_NUMBER': self.env_get_current_line_number,
                 'PLUMA_CURRENT_DOCUMENT_TYPE': self.env_get_document_type,
                 'PLUMA_DOCUMENTS_URI': self.env_get_documents_uri,
                 'PLUMA_DOCUMENTS_PATH': self.env_get_documents_path,
            }

        for var in variables:
            environ[var] = variables[var](buf)

        self.env_add_for_location(environ, buf.get_location(), 'PLUMA_CURRENT_DOCUMENT')
        return environ

    def uses_current_word(self, snippet):
        matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_WORD', snippet['text'])

        for match in matches:
            if len(match) % 2 == 0:
                return True

        return False

    def uses_current_line(self, snippet):
        matches = re.findall('(\\\\*)\\$PLUMA_CURRENT_LINE', snippet['text'])

        for match in matches:
            if len(match) % 2 == 0:
                return True

        return False

    def apply_snippet(self, snippet, start = None, end = None, environ = {}):
        if not snippet.valid:
            return False

        # Set environmental variables
        env = self.get_environment()

        if environ:
            for k in environ:
                env[k] = environ[k]

        buf = self.view.get_buffer()
        s = Snippet(snippet, env)

        if not start:
            start = buf.get_iter_at_mark(buf.get_insert())

        if not end:
            end = buf.get_iter_at_mark(buf.get_selection_bound())

        if start.equal(end) and self.uses_current_word(s):
            # There is no tab trigger and no selection and the snippet uses
            # the current word. Set start and end to the word boundary so that
            # it will be removed
            start, end = buffer_word_boundary(buf)
        elif start.equal(end) and self.uses_current_line(s):
            # There is no tab trigger and no selection and the snippet uses
            # the current line. Set start and end to the line boundary so that
            # it will be removed
            start, end = buffer_line_boundary(buf)

        # You know, we could be in an end placeholder
        (current, next) = self.next_placeholder()
        if current and current.__class__ == PlaceholderEnd:
            self.goto_placeholder(current, None)

        buf.begin_user_action()

        # Remove the tag, selection or current word
        buf.delete(start, end)

        # Insert the snippet
        if len(self.active_snippets) == 0:
            self.first_snippet_inserted()

        sn = s.insert_into(self, start)
        self.active_snippets.append(sn)

        # Put cursor at first tab placeholder
        keys = [x for x in sn.placeholders.keys() if x > 0]

        if len(keys) == 0:
            if 0 in sn.placeholders:
                self.goto_placeholder(self.active_placeholder, sn.placeholders[0])
            else:
                buf.place_cursor(sn.begin_iter())
        else:
            self.goto_placeholder(self.active_placeholder, sn.placeholders[keys[0]])

        if sn in self.active_snippets:
            # Check if we can get end_iter in view without moving the
            # current cursor position out of view
            cur = buf.get_iter_at_mark(buf.get_insert())
            last = sn.end_iter()

            curloc = self.view.get_iter_location(cur)
            lastloc = self.view.get_iter_location(last)

            if (lastloc.y + lastloc.height) - curloc.y <= \
               self.view.get_visible_rect().height:
                self.view.scroll_mark_onscreen(sn.end_mark)

        buf.end_user_action()
        self.view.grab_focus()

        return True

    def get_tab_tag(self, buf, end = None):
        if not end:
            end = buf.get_iter_at_mark(buf.get_insert())

        start = end.copy()
        word = None

        if start.backward_word_start():
            # Check if we were at a word start ourselves
            tmp = start.copy()
            tmp.forward_word_end()

            if tmp.equal(end):
                word = buf.get_text(start, end, False)
            else:
                start = end.copy()
        else:
            start = end.copy()

        if not word or word == '':
            if start.backward_char():
                word = start.get_char()

                if word.isalnum() or word.isspace():
                    return (None, None, None)
            else:
                return (None, None, None)

        return (word, start, end)

    def parse_and_run_snippet(self, data, iter):
        self.apply_snippet(DynamicSnippet(data), iter, iter)

    def run_snippet_trigger(self, trigger, bounds):
        if not self.view:
            return False

        snippets = Library().from_tag(trigger, self.language_id)
        buf = self.view.get_buffer()

        if snippets:
            if len(snippets) == 1:
                return self.apply_snippet(snippets[0], bounds[0], bounds[1])
            else:
                # Do the fancy completion dialog
                provider = Completion.Provider(_('Snippets'), self.language_id, self.on_proposal_activated)
                provider.set_proposals(snippets)

                cm = self.view.get_completion()
                cm.show([provider], cm.create_context(None))

                return True

        return False

    def run_snippet(self):
        if not self.view:
            return False

        buf = self.view.get_buffer()

        # get the word preceding the current insertion position
        (word, start, end) = self.get_tab_tag(buf)

        if not word:
            return self.skip_to_next_placeholder()

        if not self.run_snippet_trigger(word, (start, end)):
            return self.skip_to_next_placeholder()
        else:
            return True

    def deactivate_snippet(self, snippet, force = False):
        remove = []
        ordered_remove = []

        for tabstop in snippet.placeholders:
            if tabstop == -1:
                placeholders = snippet.placeholders[-1]
            else:
                placeholders = [snippet.placeholders[tabstop]]

            for placeholder in placeholders:
                if placeholder in self.placeholders:
                    if placeholder in self.update_placeholders:
                        placeholder.update_contents()

                        self.update_placeholders.remove(placeholder)
                    elif placeholder in self.jump_placeholders:
                        placeholder[0].leave()

                    remove.append(placeholder)
                elif placeholder in self.ordered_placeholders:
                    ordered_remove.append(placeholder)

        for placeholder in remove:
            if placeholder == self.active_placeholder:
                self.active_placeholder = None

            self.placeholders.remove(placeholder)
            self.ordered_placeholders.remove(placeholder)

            placeholder.remove(force)

        for placeholder in ordered_remove:
            self.ordered_placeholders.remove(placeholder)
            placeholder.remove(force)

        snippet.deactivate()
        self.active_snippets.remove(snippet)

        if len(self.active_snippets) == 0:
            self.last_snippet_removed()

        self.view.queue_draw()

    def update_snippet_contents(self):
        self.timeout_update_id = 0

        for placeholder in self.update_placeholders:
            placeholder.update_contents()

        for placeholder in self.jump_placeholders:
            self.goto_placeholder(placeholder[0], placeholder[1])

        del self.update_placeholders[:]
        del self.jump_placeholders[:]

        return False

    # Callbacks
    def on_view_destroy(self, view):
        self.stop()
        return

    def on_buffer_cursor_moved(self, buf):
        piter = buf.get_iter_at_mark(buf.get_insert())

        # Check for all snippets if the cursor is outside its scope
        for snippet in list(self.active_snippets):
            if snippet.begin_mark.get_deleted() or snippet.end_mark.get_deleted():
                self.deactivate(snippet)
            else:
                begin = snippet.begin_iter()
                end = snippet.end_iter()

                if piter.compare(begin) < 0 or piter.compare(end) > 0:
                    # Oh no! Remove the snippet this instant!!
                    self.deactivate_snippet(snippet)

        current = self.current_placeholder()

        if current != self.active_placeholder:
            self.jump_placeholders.append((self.active_placeholder, current))

            if self.timeout_update_id == 0:
                self.timeout_update_id = GLib.timeout_add(0,
                        self.update_snippet_contents)

    def on_buffer_changed(self, buf):
        for snippet in list(self.active_snippets):
            begin = snippet.begin_iter()
            end = snippet.end_iter()

            if begin.compare(end) >= 0:
                # Begin collapsed on end, just remove it
                self.deactivate_snippet(snippet)

        current = self.current_placeholder()

        if current:
            if not current in self.update_placeholders:
                self.update_placeholders.append(current)

            if self.timeout_update_id == 0:
                self.timeout_update_id = GLib.timeout_add(0, \
                        self.update_snippet_contents)

    def on_buffer_insert_text(self, buf, piter, text, length):
        ctx = get_buffer_context(buf)

        # do nothing special if there is no context and no active
        # placeholder
        if (not ctx) and (not self.active_placeholder):
            return

        if not ctx:
            ctx = self.active_placeholder

        if not ctx in self.ordered_placeholders:
            return

        # move any marks that were incorrectly moved by this insertion
        # back to where they belong
        begin = ctx.begin_iter()
        end = ctx.end_iter()
        idx = self.ordered_placeholders.index(ctx)

        for placeholder in self.ordered_placeholders:
            if placeholder == ctx:
                continue

            ob = placeholder.begin_iter()
            oe = placeholder.end_iter()

            if ob.compare(begin) == 0 and ((not oe) or oe.compare(end) == 0):
                oidx = self.ordered_placeholders.index(placeholder)

                if oidx > idx and ob:
                    buf.move_mark(placeholder.begin, end)
                elif oidx < idx and oe:
                    buf.move_mark(placeholder.end, begin)
            elif ob.compare(begin) >= 0 and ob.compare(end) < 0 and (oe and oe.compare(end) >= 0):
                buf.move_mark(placeholder.begin, end)
            elif (oe and oe.compare(begin) > 0) and ob.compare(begin) <= 0:
                buf.move_mark(placeholder.end, begin)

    def on_notify_language(self, buf, spec):
        self.update_language()

    def on_notify_editable(self, view, spec):
        self._set_view(view)

    def on_view_key_press(self, view, event):
        library = Library()

        state = event.get_state()
        keyname = Gdk.keyval_name(event.keyval)

        if not (state & Gdk.ModifierType.CONTROL_MASK) and \
                not (state & Gdk.ModifierType.MOD1_MASK) and \
                keyname in self.TAB_KEY_VAL:
            if not state & Gdk.ModifierType.SHIFT_MASK:
                return self.run_snippet()
            else:
                return self.skip_to_previous_placeholder()
        elif not library.loaded and \
                library.valid_accelerator(event.keyval, state):
            library.ensure_files()
            library.ensure(self.language_id)
            self.accelerator_activate(event.keyval, \
                    state & Gtk.accelerator_get_default_mod_mask())

        return False

    def path_split(self, path, components=[]):
        head, tail = os.path.split(path)

        if not tail and head:
            return [head] + components
        elif tail:
            return self.path_split(head, [tail] + components)
        else:
            return components

    def relative_path(self, first, second, mime):
        prot1 = re.match('(^[a-z]+:\/\/|\/)(.*)', first)
        prot2 = re.match('(^[a-z]+:\/\/|\/)(.*)', second)

        if not prot1 or not prot2:
            return second

        # Different protocols
        if prot1.group(1) != prot2.group(1):
            return second

        # Split on backslash
        path1 = self.path_split(prot1.group(2))
        path2 = self.path_split(prot2.group(2))

        # Remove as long as common
        while path1 and path2 and path1[0] == path2[0]:
            path1.pop(0)
            path2.pop(0)

        # If we need to ../ more than 3 times, then just return
        # the absolute path
        if len(path1) - 1 > 3:
            return second

        if mime.startswith('x-directory'):
            # directory, special case
            if not path2:
                result = './'
            else:
                result = '../' * (len(path1) - 1)
        else:
            # Insert ../
            result = '../' * (len(path1) - 1)

            if not path2:
                result = os.path.basename(second)

        if path2:
            result += os.path.join(*path2)

        return result

    def apply_uri_snippet(self, snippet, mime, uri):
        # Remove file scheme
        gfile = Gio.file_new_for_uri(uri)
        pathname = ''
        dirname = ''
        ruri = ''
        environ = {'PLUMA_DROP_DOCUMENT_TYPE': mime}

        self.env_add_for_location(environ, gfile, 'PLUMA_DROP_DOCUMENT')

        buf = self.view.get_buffer()
        location = buf.get_location()
        if location:
            ruri = location.get_uri()

        relpath = self.relative_path(ruri, uri, mime)

        environ['PLUMA_DROP_DOCUMENT_RELATIVE_PATH'] = relpath

        mark = buf.get_mark('gtk_drag_target')

        if not mark:
            mark = buf.get_insert()

        piter = buf.get_iter_at_mark(mark)
        self.apply_snippet(snippet, piter, piter, environ)

    def in_bounds(self, x, y):
        rect = self.view.get_visible_rect()
        rect.x, rect.y = self.view.buffer_to_window_coords(Gtk.TextWindowType.WIDGET, rect.x, rect.y)

        return not (x < rect.x or x > rect.x + rect.width or y < rect.y or y > rect.y + rect.height)

    def on_drag_data_received(self, view, context, x, y, data, info, timestamp):
        if not self.view.get_editable():
            return

        uris = drop_get_uris(data)
        if not uris:
            return

        if not self.in_bounds(x, y):
            return

        uris.reverse()
        stop = False

        for uri in uris:
            try:
                mime = Gio.content_type_guess(uri)
            except:
                mime = None

            if not mime:
                continue

            snippets = Library().from_drop_target(mime, self.language_id)

            if snippets:
                stop = True
                self.apply_uri_snippet(snippets[0], mime, uri)

        if stop:
            context.finish(True, False, timestamp)
            view.stop_emission('drag-data-received')
            view.get_toplevel().present()
            view.grab_focus()

    def find_uri_target(self, context):
        lst = Gtk.target_list_add_uri_targets((), 0)

        return self.view.drag_dest_find_target(context, lst)

    def on_proposal_activated(self, proposal, piter):
        if not self.view.get_editable():
            return False

        buf = self.view.get_buffer()
        bounds = buf.get_selection_bounds()

        if bounds:
            self.apply_snippet(proposal.snippet(), None, None)
        else:
            (word, start, end) = self.get_tab_tag(buf, piter)
            self.apply_snippet(proposal.snippet(), start, end)

        return True

    def on_default_activated(self, proposal, piter):
        buf = self.view.get_buffer()
        bounds = buf.get_selection_bounds()

        if bounds:
            buf.begin_user_action()
            buf.delete(bounds[0], bounds[1])
            buf.insert(bounds[0], proposal.props.label)
            buf.end_user_action()

            return True
        else:
            return False

    def iter_coords(self, piter):
        rect = self.view.get_iter_location(piter)
        rect.x, rect.y = self.view.buffer_to_window_coords(Gtk.TextWindowType.TEXT, rect.x, rect.y)

        return rect

    def placeholder_in_area(self, placeholder, area):
        start = placeholder.begin_iter()
        end = placeholder.end_iter()

        if not start or not end:
            return False

        # Test if start is before bottom, and end is after top
        start_rect = self.iter_coords(start)
        end_rect = self.iter_coords(end)

        return start_rect.y <= area.y + area.height and \
               end_rect.y + end_rect.height >= area.y

    def draw_placeholder_rect(self, ctx, placeholder):
        start = placeholder.begin_iter()
        start_rect = self.iter_coords(start)
        start_line = start.get_line()

        end = placeholder.end_iter()
        end_rect = self.iter_coords(end)
        end_line = end.get_line()

        line = start.copy()
        line.set_line_offset(0)
        geom = self.view.get_window(Gtk.TextWindowType.TEXT).get_geometry()

        ctx.translate(0.5, 0.5)

        while line.get_line() <= end_line:
            ypos, height = self.view.get_line_yrange(line)
            x_, ypos = self.view.window_to_buffer_coords(Gtk.TextWindowType.TEXT, 0, ypos)

            if line.get_line() == start_line and line.get_line() == end_line:
                # Simply draw a box, both are on the same line
                ctx.rectangle(start_rect.x, start_rect.y, end_rect.x - start_rect.x, start_rect.height - 1)
                ctx.stroke()
            elif line.get_line() == start_line or line.get_line() == end_line:
                if line.get_line() == start_line:
                    rect = start_rect
                else:
                    rect = end_rect

                ctx.move_to(0, rect.y + rect.height - 1)
                ctx.rel_line_to(rect.x, 0)
                ctx.rel_line_to(0, -rect.height + 1)
                ctx.rel_line_to(geom[2], 0)
                ctx.stroke()

            if not line.forward_line():
                break

    def draw_placeholder_bar(self, ctx, placeholder):
        start = placeholder.begin_iter()
        start_rect = self.iter_coords(start)

        ctx.translate(0.5, 0.5)
        extend_width = 2.5

        ctx.move_to(start_rect.x - extend_width, start_rect.y)
        ctx.rel_line_to(extend_width * 2, 0)

        ctx.move_to(start_rect.x, start_rect.y)
        ctx.rel_line_to(0, start_rect.height - 1)

        ctx.rel_move_to(-extend_width, 0)
        ctx.rel_line_to(extend_width * 2, 0)
        ctx.stroke()

    def draw_placeholder(self, ctx, placeholder):
        if isinstance(placeholder, PlaceholderEnd):
            return

        col = self.view.get_style_context().get_color(Gtk.StateFlags.INSENSITIVE)
        col.alpha = 0.5
        Gdk.cairo_set_source_rgba(ctx, col)

        if placeholder.tabstop > 0:
            ctx.set_dash([], 0)
        else:
            ctx.set_dash([2], 0)

        start = placeholder.begin_iter()
        end = placeholder.end_iter()

        if start.equal(end):
            self.draw_placeholder_bar(ctx, placeholder)
        else:
            self.draw_placeholder_rect(ctx, placeholder)

    def on_draw(self, view, ctx):
        window = view.get_window(Gtk.TextWindowType.TEXT)

        if not Gtk.cairo_should_draw_window(ctx, window):
            return False

        ctx.set_line_width(1.0)
        Gtk.cairo_transform_to_window(ctx, view, window)
        clipped, clip = Gdk.cairo_get_clip_rectangle(ctx)

        if not clipped:
            return False

        for placeholder in self.ordered_placeholders:
            if not self.placeholder_in_area(placeholder, clip):
                continue

            ctx.save()
            self.draw_placeholder(ctx, placeholder)
            ctx.restore()

        return False

# ex:ts=4:et:



#    Pluma snippets plugin
#    Copyright (C) 2005-2006  Jesse van den Kieboom <jesse@icecrew.nl>
#    Copyright (C) 2012-2021 MATE Developers
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import six
from gi.repository import Gio, Gtk

from .Placeholder import *
from .Parser import Parser, Token
from .Helper import *

class EvalUtilities:
    def __init__(self, view=None):
        self.view = view
        self._init_namespace()

    def _init_namespace(self):
        self.namespace = {
            '__builtins__': __builtins__,
            'align': self.util_align,
            'readfile': self.util_readfile,
            'filesize': self.util_filesize
        }

    def _real_len(self, s, tablen = 0):
        if tablen == 0:
            tablen = self.view.get_tab_width()

        return len(s.expandtabs(tablen))

    def _filename_to_uri(self, filename):
        gfile = Gio.file_new_for_path(filename)

        return gfile.get_uri()

    def util_readfile(self, filename):
        stream = Gio.file_new_for_path(filename).read()

        if not stream:
            return ''

        res = stream.read()
        stream.close()

        return res

    def util_filesize(self, filename):
        gfile = Gio.file_new_for_path(filename)
        info = gfile.query_info(Gio.FILE_ATTRIBUTE_STANDARD_SIZE)

        if not info:
            return 0

        return info.get_size()

    def util_align(self, items):
        maxlen = []
        tablen = self.view.get_tab_width()

        for row in range(0, len(items)):
            for col in range(0, len(items[row]) - 1):
                if row == 0:
                    maxlen.append(0)

                items[row][col] += "\t"
                rl = self._real_len(items[row][col], tablen)

                if (rl > maxlen[col]):
                    maxlen[col] = rl

        result = ''

        for row in range(0, len(items)):
            for col in range(0, len(items[row]) - 1):
                item = items[row][col]

                result += item + ("\t" * int(((maxlen[col] - \
                        self._real_len(item, tablen)) / tablen)))

            result += items[row][len(items[row]) - 1]

            if row != len(items) - 1:
                result += "\n"

        return result

class Snippet:
    def __init__(self, data, environ = {}):
        self.data = data
        self.environ = environ

    def __getitem__(self, prop):
        return self.data[prop]

    def __setitem__(self, prop, value):
        self.data[prop] = value

    def accelerator_display(self):
        accel = self['accelerator']

        if accel:
            keyval, mod = Gtk.accelerator_parse(accel)
            accel = Gtk.accelerator_get_label(keyval, mod)

        return accel or ''

    def display(self):
        nm = markup_escape(self['description'])

        tag = self['tag']
        accel = self.accelerator_display()
        detail = []

        if tag and tag != '':
            detail.append(tag)

        if accel and accel != '':
            detail.append(accel)

        if not detail:
            return nm
        else:
            return nm + ' (<b>' + markup_escape(", ".join(detail)) + \
                    '</b>)'

    def _add_placeholder(self, placeholder):
        if placeholder.tabstop in self.placeholders:
            if placeholder.tabstop == -1:
                self.placeholders[-1].append(placeholder)
                self.plugin_data.ordered_placeholders.append(placeholder)
        elif placeholder.tabstop == -1:
            self.placeholders[-1] = [placeholder]
            self.plugin_data.ordered_placeholders.append(placeholder)
        else:
            self.placeholders[placeholder.tabstop] = placeholder
            self.plugin_data.ordered_placeholders.append(placeholder)

    def _insert_text(self, text):
        # Insert text keeping indentation in mind
        indented = (six.u('\n') + self._indent).join(spaces_instead_of_tabs(self._view, text).split('\n'))
        self._view.get_buffer().insert(self._insert_iter(), indented)

    def _insert_iter(self):
        return self._view.get_buffer().get_iter_at_mark(self._insert_mark)

    def _create_environment(self, data):
        if data in self.environ:
            val = self.environ[data]
        else:
            val = ''

        # Get all the current indentation
        all_indent = compute_indentation(self._view, self._insert_iter())

        # Substract initial indentation to get the snippet indentation
        indent = all_indent[len(self._indent):]

        # Keep indentation
        return (six.u('\n') + indent).join(val.split('\n'))

    def _create_placeholder(self, data):
        tabstop = data['tabstop']
        begin = self._insert_iter()

        if tabstop == 0:
            # End placeholder
            return PlaceholderEnd(self._view, self.environ, begin, data['default'])
        elif tabstop in self.placeholders:
            # Mirror placeholder
            return PlaceholderMirror(self._view, tabstop, self.environ, begin)
        else:
            # Default placeholder
            return Placeholder(self._view, tabstop, self.environ, data['default'], begin)

    def _create_shell(self, data):
        begin = self._insert_iter()
        return PlaceholderShell(self._view, data['tabstop'], self.environ, begin, data['contents'])

    def _create_eval(self, data):
        begin = self._insert_iter()
        return PlaceholderEval(self._view, data['tabstop'], self.environ, data['dependencies'], begin, data['contents'], self._utils.namespace)

    def _create_regex(self, data):
        begin = self._insert_iter()
        return PlaceholderRegex(self._view, data['tabstop'], self.environ, begin, data['input'], data['pattern'], data['substitution'], data['modifiers'])

    def _create_text(self, data):
        return data

    def _invalid_placeholder(self, placeholder, remove):
        buf = self._view.get_buffer()

        # Remove the text because this placeholder is invalid
        if placeholder.default and remove:
            buf.delete(placeholder.begin_iter(), placeholder.end_iter())

        placeholder.remove()

        if placeholder.tabstop == -1:
            index = self.placeholders[-1].index(placeholder)
            del self.placeholders[-1][index]
        else:
            del self.placeholders[placeholder.tabstop]

        self.plugin_data.ordered_placeholders.remove(placeholder)

    def _parse(self, plugin_data):
        # Initialize current variables
        self._view = plugin_data.view
        self._indent = compute_indentation(self._view, self._view.get_buffer().get_iter_at_mark(self.begin_mark))
        self._utils = EvalUtilities(self._view)
        self.placeholders = {}
        self._insert_mark = self.end_mark
        self.plugin_data = plugin_data

        # Create parser
        parser = Parser(data=self['text'])

        # Parse tokens
        while (True):
            token = parser.token()

            if not token:
                break

            try:
                val = {'environment': self._create_environment,
                    'placeholder': self._create_placeholder,
                    'shell': self._create_shell,
                    'eval': self._create_eval,
                    'regex': self._create_regex,
                    'text': self._create_text}[token.klass](token.data)
            except KeyError:
                sys.stderr.write('Token class not supported: %s\n' % token.klass)
                continue

            if isinstance(val, six.string_types):
                # Insert text
                self._insert_text(val)
            else:
                # Insert placeholder
                self._add_placeholder(val)

        # Create end placeholder if there isn't one yet
        if 0 not in self.placeholders:
            self.placeholders[0] = PlaceholderEnd(self._view, self.environ, self.end_iter(), None)
            self.plugin_data.ordered_placeholders.append(self.placeholders[0])

        # Make sure run_last is ran for all placeholders and remove any
        # non `ok` placeholders
        for tabstop in self.placeholders.copy():
            ph = (tabstop == -1 and list(self.placeholders[-1])) or [self.placeholders[tabstop]]

            for placeholder in ph:
                placeholder.run_last(self.placeholders)

                if not placeholder.ok or placeholder.done:
                    self._invalid_placeholder(placeholder, not placeholder.ok)

        # Remove all the Expand placeholders which have a tabstop because
        # they can be used to mirror, but they shouldn't be real tabstops
        # (if they have mirrors installed). This is problably a bit of
        # a dirty hack :)
        if -1 not in self.placeholders:
            self.placeholders[-1] = []

        for tabstop in self.placeholders.copy():
            placeholder = self.placeholders[tabstop]

            if tabstop != -1:
                if isinstance(placeholder, PlaceholderExpand) and \
                   placeholder.has_references:
                    # Add to anonymous placeholders
                    self.placeholders[-1].append(placeholder)

                    # Remove placeholder
                    del self.placeholders[tabstop]

        self.plugin_data = None

    def insert_into(self, plugin_data, insert):
        buf = plugin_data.view.get_buffer()
        last_index = 0

        # Find closest mark at current insertion, so that we may insert
        # our marks in the correct order
        (current, next) = plugin_data.next_placeholder()

        if current:
            # Insert AFTER current
            last_index = plugin_data.placeholders.index(current) + 1
        elif next:
            # Insert BEFORE next
            last_index = plugin_data.placeholders.index(next)
        else:
            # Insert at first position
            last_index = 0

        # lastIndex now contains the position of the last mark
        # Create snippet bounding marks
        self.begin_mark = buf.create_mark(None, insert, True)
        self.end_mark = buf.create_mark(None, insert, False)

        # Now parse the contents of this snippet, create Placeholders
        # and insert the placholder marks in the marks array of plugin_data
        self._parse(plugin_data)

        # So now all of the snippet is in the buffer, we have all our
        # placeholders right here, what's next, put all marks in the
        # plugin_data.marks
        k = sorted(self.placeholders.keys(), reverse=True)

        plugin_data.placeholders.insert(last_index, self.placeholders[0])
        last_iter = self.placeholders[0].end_iter()

        for tabstop in k:
            if tabstop != -1 and tabstop != 0:
                placeholder = self.placeholders[tabstop]
                end_iter = placeholder.end_iter()

                if last_iter.compare(end_iter) < 0:
                    last_iter = end_iter

                # Inserting placeholder
                plugin_data.placeholders.insert(last_index, placeholder)

        # Move end mark to last placeholder
        buf.move_mark(self.end_mark, last_iter)

        return self

    def deactivate(self):
        buf = self.begin_mark.get_buffer()

        buf.delete_mark(self.begin_mark)
        buf.delete_mark(self.end_mark)

        self.placeholders = {}

    def begin_iter(self):
        return self.begin_mark.get_buffer().get_iter_at_mark(self.begin_mark)

    def end_iter(self):
        return self.end_mark.get_buffer().get_iter_at_mark(self.end_mark)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2009 - Jesse van den Kieboom
#  Copyright (C) 2012-2021 MATE Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os
import codecs
from gi.repository import Gio, GLib, Gtk, Pluma
from .popup import Popup
from .virtualdirs import RecentDocumentsDirectory
from .virtualdirs import CurrentDocumentsDirectory

ui_str = """<ui>
  <menubar name="MenuBar">
    <menu name="FileMenu" action="File">
      <placeholder name="FileOps_2">
    <menuitem name="QuickOpen" action="QuickOpen"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class WindowHelper:
    def __init__(self, window, plugin):
        self._window = window
        self._plugin = plugin

        self._popup = None
        self._install_menu()

    def deactivate(self):
        self._uninstall_menu()
        self._window = None
        self._plugin = None

    def update_ui(self):
        pass

    def _uninstall_menu(self):
        manager = self._window.get_ui_manager()

        manager.remove_ui(self._ui_id)
        manager.remove_action_group(self._action_group)

        manager.ensure_update()

    def _install_menu(self):
        manager = self._window.get_ui_manager()
        action = Gtk.Action.new("QuickOpen",
                                _("Quick open"),
                                _("Quickly open documents"))
        action.set_icon_name("document-open")
        action.connect("activate", self.on_quick_open_activate)
        self._action_group = Gtk.ActionGroup("PlumaQuickOpenPluginActions")
        self._action_group.add_action_with_accel(action, "<Ctrl><Alt>O")

        manager.insert_action_group(self._action_group, -1)
        self._ui_id = manager.add_ui_from_string(ui_str)

    def _create_popup(self):
        paths = []

        # Open documents
        paths.append(CurrentDocumentsDirectory(self._window))

        doc = self._window.get_active_document()

        # Current document directory
        if doc and doc.is_local():
            gfile = doc.get_location()
            paths.append(gfile.get_parent())

        # File browser root directory
        bus = self._window.get_message_bus()

        try:
            msg = bus.send_sync('/plugins/filebrowser', 'get_root')

            if msg:
                uri = msg.get_value('uri')

                if uri:
                    gfile = Gio.file_new_for_uri(uri)

                    if gfile and gfile.is_native():
                        paths.append(gfile)

        except Exception:
            pass

        # Recent documents
        paths.append(RecentDocumentsDirectory())

        # Local bookmarks
        for path in self._local_bookmarks():
            paths.append(path)

        # Desktop directory
        desktopdir = self._desktop_dir()

        if desktopdir:
            paths.append(Gio.file_new_for_path(desktopdir))

        # Home directory
        paths.append(Gio.file_new_for_path(os.path.expanduser('~')))

        self._popup = Popup(self._window, paths, self.on_activated)

        self._popup.set_default_size(*self._plugin.get_popup_size())
        self._popup.set_transient_for(self._window)
        self._popup.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        self._window.get_group().add_window(self._popup)

        self._popup.connect('destroy', self.on_popup_destroy)

    def _local_bookmarks(self):
        filename = os.path.expanduser('~/.config/gtk-3.0/bookmarks')

        if not os.path.isfile(filename):
            return []

        paths = []

        for line in codecs.open(filename, 'r', encoding='utf-8'):
            uri = line.strip().split(" ")[0]
            f = Gio.file_new_for_uri(uri)

            if f.is_native():
                try:
                    info = f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE, None)

                    if info and info.get_file_type() == Gio.FileType.DIRECTORY:
                        paths.append(f)
                except GLib.GError:
                    pass

        return paths

    def _desktop_dir(self):
        config = os.getenv('XDG_CONFIG_HOME')

        if not config:
            config = os.path.expanduser('~/.config')

        config = os.path.join(config, 'user-dirs.dirs')
        desktopdir = None

        if os.path.isfile(config):
            for line in codecs.open(config, 'r', encoding='utf-8'):
                line = line.strip()

                if line.startswith('XDG_DESKTOP_DIR'):
                    parts = line.split('=', 1)
                    desktopdir = os.path.expandvars(parts[1].strip('"').strip("'"))
                    break

        if not desktopdir:
            desktopdir = os.path.expanduser('~/Desktop')

        return desktopdir

    # Callbacks
    def on_quick_open_activate(self, action):
        if not self._popup:
            self._create_popup()

        self._popup.show()

    def on_popup_destroy(self, popup):
        self._plugin.set_popup_size(popup.get_final_size())
        self._popup = None

    def on_activated(self, gfile):
        Pluma.commands_load_uri(self._window, gfile.get_uri(), None, -1)
        return True

# ex:ts=4:et:



import os
from gi.repository import GtkSource

from .Library import Library

global manager
manager = None

def get_language_manager():
    global manager

    if not manager:
        dirs = []

        for d in Library().systemdirs:
            dirs.append(os.path.join(d, 'lang'))

        manager = GtkSource.LanguageManager()
        manager.set_search_path(dirs + manager.get_search_path())

    return manager

# ex:ts=4:et:



import os
import tempfile
import sys
import shutil

from .Library import *
import xml.etree.ElementTree as et
from .Helper import *

class Exporter:
    def __init__(self, filename, snippets):
        self.filename = filename
        self.set_snippets(snippets)

    def set_snippets(self, snippets):
        self.snippets = {}

        for snippet in snippets:
            lang = snippet.language()

            if lang in self.snippets:
                self.snippets[lang].append(snippet)
            else:
                self.snippets[lang] = [snippet]

    def export_xml(self, dirname, language, snippets):
        # Create the root snippets node
        root = et.Element('snippets')

        # Create filename based on language
        if language:
            filename = os.path.join(dirname, language + '.xml')

            # Set the language attribute
            root.attrib['language'] = language
        else:
            filename = os.path.join(dirname, 'global.xml')

        # Add all snippets to the root node
        for snippet in snippets:
            root.append(snippet.to_xml())

        # Write xml
        write_xml(root, filename, ('text', 'accelerator'))

    def export_archive(self, cmd):
        dirname = tempfile.mkdtemp()

        # Save current working directory and change to temporary directory
        curdir = os.getcwd()

        try:
            os.chdir(dirname)

            # Write snippet xml files
            for language, snippets in self.snippets.items():
                self.export_xml(dirname, language , snippets)

            # Archive files
            status = os.system('%s "%s" *.xml' % (cmd, self.filename))
        finally:
            os.chdir(curdir)

        if status != 0:
            return _('The archive "%s" could not be created' % self.filename)

        # Remove the temporary directory
        shutil.rmtree(dirname)

    def export_targz(self):
        self.export_archive('tar -c --gzip -f')

    def export_tarbz2(self):
        self.export_archive('tar -c --bzip2 -f')

    def export_tar(self):
        self.export_archive('tar -cf')

    def run(self):
        dirname = os.path.dirname(self.filename)
        if not os.path.exists(dirname):
            return _('Target directory "%s" does not exist') % dirname

        if not os.path.isdir(dirname):
            return _('Target directory "%s" is not a valid directory') % dirname

        (root, ext) = os.path.splitext(self.filename)

        actions = {'.tar.gz': self.export_targz,
               '.tar.bz2': self.export_tarbz2,
               '.tar': self.export_tar}

        for k, v in actions.items():
            if self.filename.endswith(k):
                return v()

        return self.export_targz()
# ex:ts=4:et:



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# generate-plugin.py - pluma plugin skeletton generator
# This file is part of pluma
#
# Copyright (C) 2006 - Steve Frécinaux
# Copyright (C) 2012-2021 MATE Developers
#
# pluma is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pluma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pluma; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA  02110-1301  USA

import re
import os
import sys
import getopt
from datetime import date
import preprocessor

# Default values of command line options
options = {
    'language'          : 'c',
    'description'       : 'Type here a short description of your plugin',
    'author'            : os.getenv('USERNAME'),
    'email'             : os.getenv('LOGNAME') + '@email.com',
    'standalone'        : False,
    'with-side-pane'    : False,
    'with-bottom-pane'  : False,
    'with-menu'         : False,
    'with-config-dlg'   : False
}

USAGE = """Usage:
   %s [OPTIONS...] pluginname
""" % os.path.basename(sys.argv[0])
HELP = USAGE + """
generate skeleton source tree for a new pluma plugin.

Options:
  --author              Set the author name
  --email               Set the author email
  --description         Set the description you want for your new plugin
  --standalone          Is this plugin intended to be distributed as a
                        standalone package ? (N/A)
  --language / -l       Set the language (C) [default: %(language)s]
  --with-$feature       Enable $feature
  --without-$feature    Disable $feature
  --help / -h           Show this message and exits

Features:
  config-dlg            Plugin configuration dialog
  menu                  Plugin menu entries
  side-pane             Side pane item (N/A)
  bottom-pane           Bottom pane item (N/A)
""" % options

TEMPLATE_DIR = os.path.join(os.path.dirname(sys.argv[0]), "plugin_template")

# Parsing command line options
try:
    opts, args = getopt.getopt(sys.argv[1:],
                               'l:h',
                               ['language=',
                                'description=',
                                'author=',
                                'email=',
                                'standalone',
                                'with-menu'         , 'without-menu',
                                'with-side-pane'    , 'without-side-pane',
                                'with-bottom-pane'  , 'without-bottom-pane',
                                'with-config-dlg'   , 'without-config-dlg',
                                'help'])
except getopt.error, exc:
    print >>sys.stderr, '%s: %s' % (sys.argv[0], str(exc))
    print >>sys.stderr, USAGE
    sys.exit(1)

for opt, arg in opts:
    if opt in ('-h', '--help'):
        print >>sys.stderr, HELP
        sys.exit(0)

    elif opt in ('--description', '--author', '--email'):
        options[opt[2:]] = arg

    elif opt in ('-l', '--language'):
        options['language'] = arg.lower()

    elif opt == '--standalone':
        options['standalone'] = True

    elif opt[0:7] == '--with-':
        options['with-' + opt[7:]] = True

    elif opt[0:10] == '--without-':
        options['with-' + opt[10:]] = False

# What's the new plugin name ?
if len(args) < 1:
    print >>sys.stderr, USAGE
    sys.exit(1)

plugin_name = args[0]
plugin_id = re.sub('[^a-z0-9_]', '', plugin_name.lower().replace(' ', '_'))
plugin_module = plugin_id.replace('_', '-')

directives = {
    'PLUGIN_NAME'       : plugin_name,
    'PLUGIN_MODULE'     : plugin_module,
    'PLUGIN_ID'         : plugin_id,
    'AUTHOR_FULLNAME'   : options['author'],
    'AUTHOR_EMAIL'      : options['email'],
    'DATE_YEAR'         : date.today().year,
    'DESCRIPTION'       : options['description'],
}

# Files to be generated by the preprocessor, in the form "template : outfile"
output_files = {
    'Makefile.am': '%s/Makefile.am' % plugin_module,
    'pluma-plugin.desktop.in': '%s/%s.pluma-plugin.desktop.in' % (plugin_module, plugin_module)
}

if options['language'] == 'c':
    output_files['pluma-plugin.c'] = '%s/%s-plugin.c' % (plugin_module, plugin_module)
    output_files['pluma-plugin.h'] = '%s/%s-plugin.h' % (plugin_module, plugin_module)
else:
    print >>sys.stderr, 'Value of --language should be C'
    print >>sys.stderr, USAGE
    sys.exit(1)

if options['standalone']:
    output_files['configure.ac'] = 'configure.ac'

if options['with-side-pane']:
    directives['WITH_SIDE_PANE'] = True

if options['with-bottom-pane']:
    directives['WITH_BOTTOM_PANE'] = True

if options['with-menu']:
    directives['WITH_MENU'] = True

if options['with-config-dlg']:
    directives['WITH_CONFIGURE_DIALOG'] = True


# Generate the plugin base
for infile, outfile in output_files.iteritems():
    print 'Processing %s\n' \
          '      into %s...' % (infile, outfile)

    infile = os.path.join(TEMPLATE_DIR, infile)
    outfile = os.path.join(os.getcwd(), outfile)

    if not os.path.isfile(infile):
        print >>sys.stderr, 'Input file does not exist : %s.' % os.path.basename(infile)
        continue

    # Make sure the destination directory exists
    if not os.path.isdir(os.path.split(outfile)[0]):
        os.makedirs(os.path.split(outfile)[0])

    # Variables relative to the generated file
    directives['DIRNAME'], directives['FILENAME'] = os.path.split(outfile)

    # Generate the file
    preprocessor.process(infile, outfile, directives.copy())

print 'Done.'

# ex:ts=4:et:
